"""
Interop PyCUDA
~~~~~~~~~~~~~~
"""
from bohrium_api import stack_info
from .bhary import get_base
from ._bh import get_data_pointer
from .backend_messaging import cuda_use_current_context
from . import contexts


_iniziated = False


def _import_pycuda_module():
    """Help function to import PyCUDA and checks that a CUDA backend is present"""
    try:
        import pycuda
    except ImportError:
        raise ImportError("Failed to import the `pycuda` module, please install PyCUDA")

    if not stack_info.is_cuda_in_stack():
        raise RuntimeError("No CUDA device in the Bohrium stack! "
                           "Try defining the environment variable `BH_STACK=cuda`.")

    if stack_info.is_proxy_in_stack():
        raise RuntimeError("Cannot directly access the CUDA device through a proxy.")
    return pycuda


def available():
    """Is CUDA available?"""
    try:
        _import_pycuda_module()
        return True
    except ImportError:
        return False
    except RuntimeError:
        return False


def init():
    """Initiate the PyCUDA module. Must be called before any other PyCUDA calls and
    preferable also before any Bohrium operations"""
    global _iniziated
    if _iniziated:
        return
    _initiated = True
    if not available():
        return
    from . import _bh
    _bh.flush()
    import pycuda
    import pycuda.autoinit
    pycuda.driver.mem_alloc(1)  # Force PyCUDA to activate the context as "current context"
    cuda_use_current_context()  # And then tell Bohrium to use that context


def get_gpuarray(bh_ary):
    """Return a PyCUDA GPUArray object that points to the same device memory as `bh_ary`.

    Parameters
    ----------
    bh_ary : ndarray (Bohrium array)
        Must be a Bohrium base array

    Returns
    -------
    out : GPUArray

    Notes
    -----
    Changing or deallocating `bh_ary` invalidates the returned GPUArray array!

    """

    if get_base(bh_ary) is not bh_ary:
        raise RuntimeError('`bh_ary` must be a base array and not a view')
    assert (bh_ary.bhc_mmap_allocated)

    with contexts.DisableBohrium():
        _import_pycuda_module()
        from pycuda import gpuarray
        dev_ptr = get_data_pointer(get_base(bh_ary), copy2host=False, allocate=True)
        return gpuarray.GPUArray(bh_ary.shape, bh_ary.dtype, gpudata=dev_ptr)


def max_local_memory(cuda_device=None):
    """Returns the maximum allowed local memory (memory per block) on `cuda_device`.
       If `cuda_device` is None, use current device"""

    pycuda = _import_pycuda_module()
    if cuda_device is None:
        cuda_device = pycuda.driver.Context.get_device()
    return cuda_device.get_attributes()[pycuda.driver.device_attribute.MAX_SHARED_MEMORY_PER_BLOCK]


def type_np2cuda_str(np_type):
    """Converts a NumPy type to a CUDA type string"""
    import numpy as np
    if np_type == np.bool:
        return "bool"
    elif np_type == np.int8:
        return "char"
    elif np_type == np.int16:
        return "short"
    if np_type == np.int32:
        return "int"
    elif np_type == np.int64:
        return "long"
    elif np_type == np.uint8:
        return "uchar"
    elif np_type == np.uint16:
        return "ushort"
    elif np_type == np.uint32:
        return "uint"
    elif np_type == np.uint64:
        return "ulong"
    elif np_type == np.float32:
        return "float"
    elif np_type == np.float64:
        return "double"
    else:
        return "UNKNOWN"



import numpy_force as np
from bohrium_api import stack_info
from . import _bh, bhary, _util, array_create

_default_compiler_command = None


def get_default_compiler_command():
    """ Returns the default compiler command, which is the one typically extended with extra link commands """
    global _default_compiler_command
    if _default_compiler_command is None:
        import re
        from bohrium_api import stack_info
        m = re.search("JIT Command: \"([^\"]*)\"", stack_info.info()['runtime_info'])
        if m is None:
            raise RuntimeError("'JIT Command' not found in the Bohrium backend")
        _default_compiler_command = m.group(1)
    return _default_compiler_command


def execute(kernel_source, operand_list, compiler_command=None, tag="openmp", param=None, only_behaving_operands=True):
    """ Compile and execute the function `execute()` with the arguments in `operand_list`

    Parameters
    ----------
    kernel_source : str
        The kernel source code that most define the function `execute()` that should take arguments corresponding
        to the `operand_list`
    operand_list : list of bohrium arrays
        The arrays given to the `execute()` function defined in `kernel_source`
    compiler_command : str, optional
        The compiler command to use when comping the kernel. `{OUT}` and `{IN}` in the command are replaced with the
        name of the binary and source path.
        When this options isn't specified, the default command are used see `get_default_compiler_command()`.
    tag : str, optional
        Name of the backend that should handle this kernel.
    param : dict, optional
        Backend specific parameters (e.g. OpenCL needs `global_work_size` and `local_work_size`).
    only_behaving_operands : bool, optional
        Set to False in order to allow non-behaving operands. Requirements for a behaving array:
             * Is a bohrium array
             * Is C-style contiguous
             * Points to the first element in the underlying base array (no offset)
             * Has the same total length as its base
        See `make_behaving()`

    Examples
    --------
    # Simple addition kernel
    import bohrium as bh
    kernel = r'''
    #include <stdint.h>
    void execute(double *a, double *b, double *c) {
        for(uint64_t i=0; i<100; ++i) {
            c[i] = a[i] + b[i] + i;
        }
    }'''
    a = bh.ones(100, bh.double)
    b = bh.ones(100, bh.double)
    res = bh.empty_like(a)
    bh.user_kernel.execute(kernel, [a, b, res])
    """
    if stack_info.is_proxy_in_stack():
        raise RuntimeError("The proxy backend does not support user kernels")
    if compiler_command is None:
        compiler_command = get_default_compiler_command()
    for op in operand_list:
        if not bhary.check(op):
            raise TypeError("All operands in `operand_list` must be Bohrium arrays")
        if only_behaving_operands and not _bh.is_array_behaving(op):
            raise TypeError("Operand is not behaving set `only_behaving_operands=False` or use `make_behaving()`")

    if param is None:
        param = {}

    def parse_param():
        import collections
        for key, value in param.items():
            if isinstance(value, collections.Iterable) and not isinstance(value, str):
                value = " ".join(str(subvalue) for subvalue in value)
            yield "%s: %s" % (key, value)

    param_str = "; ".join(parse_param())

    _bh.flush()
    ret_msg = _bh.user_kernel(kernel_source, operand_list, compiler_command, tag, param_str)
    if len(ret_msg) > 0:
        raise RuntimeError(ret_msg)


def dtype_to_c99(dtype):
    """ Returns the C99 name of `dtype` """
    if np.issubdtype(dtype, np.integer):
        return "%s_t" % str(dtype)
    elif _util.dtype_equal(dtype, np.float32):
        return "float"
    elif _util.dtype_equal(dtype, np.float64):
        return "double"
    elif _util.dtype_equal(dtype, np.complex64):
        return "float complex"
    elif _util.dtype_equal(dtype, np.complex128):
        return "double complex"
    raise TypeError("dtype '%s' unsupported" % str(dtype))


def gen_function_prototype(operand_list, operand_name_list=None):
    """ Returns the `execute() definition based on the arrays in `operand_list` """
    dtype_list = [dtype_to_c99(t.dtype) for t in operand_list]
    ret = "#include <stdint.h>\n#include <complex.h>\n"
    ret += "void execute("
    for i in range(len(dtype_list)):
        ret += "%s *" % dtype_list[i]
        if operand_name_list is None:
            ret += "a%d, " % i
        else:
            ret += "%s, " % operand_name_list[i]
    return "%s)\n" % ret[:-2]


def make_behaving(ary, dtype=None):
    """ Make sure that `ary` is a "behaving" bohrium array of type `dtype`.
    Requirements for a behaving array:
     * Is a bohrium array
     * Is C-style contiguous
     * Points to the first element in the underlying base array (no offset)
     * Has the same total length as its base

    Parameters
    ----------
    ary : array_like
        The array to make behaving
    dtype : boolean, optional
        The return array is converted to `dtype` if not None
    
    Returns
    -------
    A behaving Bohrium array that might be a copy of `ary`

    Note
    ----
    Use this function to make sure that operands given to `execute()` is "behaving" that is
    the kernel can access the arrays without worrying about offset and stride.
    """

    ary = array_create.array(ary, dtype=dtype, order='C')
    if not _bh.is_array_behaving(ary):
        ary = array_create.array(ary, copy=True)
    return ary



"""
Interop PyOpenCL
~~~~~~~~~~~~~~~~
"""
from bohrium_api import stack_info
from .bhary import get_base
from ._bh import get_data_pointer, set_data_pointer, get_device_context


def _import_pyopencl_module():
    """Help function to import PyOpenCL and checks that a OpenCL backend is present"""
    try:
        import pyopencl
    except ImportError:
        raise ImportError("Failed to import the `pyopencl` module, please install PyOpenCL")

    if not stack_info.is_opencl_in_stack():
        raise RuntimeError("No OpenCL device in the Bohrium stack! "
                           "Try defining the environment variable `BH_STACK=opencl`.")

    if stack_info.is_proxy_in_stack():
        raise RuntimeError("Cannot directly access the OpenCL device through a proxy.")
    return pyopencl


def available():
    """Is PyOpenCL available?"""
    try:
        _import_pyopencl_module()
        return True
    except ImportError:
        return False
    except RuntimeError:
        return False


def get_context():
    """Return a PyOpenCL context"""
    pyopencl = _import_pyopencl_module()
    cxt = get_device_context()
    if cxt is None:
        raise RuntimeError("No OpenCL device in the Bohrium stack! "
                           "Try defining the environment variable `BH_STACK=opencl`.")
    return pyopencl.Context.from_int_ptr(cxt)


def get_buffer(bh_ary):
    """Return a OpenCL Buffer object wrapping the Bohrium array `ary`.

    Parameters
    ----------
    bh_ary : ndarray (Bohrium array)
        Must be a Bohrium base array

    Returns
    -------
    out : pyopencl.Buffer

    Notes
    -----
    Changing or deallocating `bh_ary` invalidates the returned pyopencl.Buffer!

    """

    if get_base(bh_ary) is not bh_ary:
        raise RuntimeError('`bh_ary` must be a base array and not a view')
    assert (bh_ary.bhc_mmap_allocated)

    pyopencl = _import_pyopencl_module()
    cl_mem = get_data_pointer(get_base(bh_ary), copy2host=False, allocate=True)
    return pyopencl.Buffer.from_int_ptr(cl_mem)


def set_buffer(bh_ary, buffer):
    """Assign a OpenCL Buffer object to a Bohrium array `ary`.

    Parameters
    ----------
    bh_ary : ndarray (Bohrium array)
        Must be a Bohrium base array

    buffer : pyopencl.Buffer
        The PyOpenCL device buffer
    """

    if get_base(bh_ary) is not bh_ary:
        raise RuntimeError('`bh_ary` must be a base array and not a view')

    set_data_pointer(get_base(bh_ary), buffer.int_ptr, host_ptr=False)


def get_array(bh_ary, queue):
    _import_pyopencl_module()
    from pyopencl import array as clarray
    return clarray.Array(queue, bh_ary.shape, bh_ary.dtype, data=get_buffer(bh_ary))


def kernel_info(opencl_kernel, queue):
    """Info about the `opencl_kernel`
    Returns 4-tuple:
            - Max work-group size
            - Recommended work-group multiple
            - Local mem used by kernel
            - Private mem used by kernel
    """
    cl = _import_pyopencl_module()
    info = cl.kernel_work_group_info
    # Max work-group size
    wg_size = opencl_kernel.get_work_group_info(info.WORK_GROUP_SIZE, queue.device)
    # Recommended work-group multiple
    wg_multiple = opencl_kernel.get_work_group_info(info.PREFERRED_WORK_GROUP_SIZE_MULTIPLE, queue.device)
    # Local mem used by kernel
    local_usage = opencl_kernel.get_work_group_info(info.LOCAL_MEM_SIZE, queue.device)
    # Private mem used by kernel
    private_usage = opencl_kernel.get_work_group_info(info.PRIVATE_MEM_SIZE, queue.device)
    return (wg_size, wg_multiple, local_usage, private_usage)


def max_local_memory(opencl_device):
    """Returns the maximum allowed local memory on `opencl_device`"""
    cl = _import_pyopencl_module()
    return opencl_device.get_info(cl.device_info.LOCAL_MEM_SIZE)


def type_np2opencl_str(np_type):
    """Converts a NumPy type to a OpenCL type string"""
    import numpy as np
    if np_type == np.bool:
        return "bool"
    elif np_type == np.int8:
        return "char"
    elif np_type == np.int16:
        return "short"
    if np_type == np.int32:
        return "int"
    elif np_type == np.int64:
        return "long"
    elif np_type == np.uint8:
        return "uchar"
    elif np_type == np.uint16:
        return "ushort"
    elif np_type == np.uint32:
        return "uint"
    elif np_type == np.uint64:
        return "ulong"
    elif np_type == np.float32:
        return "float"
    elif np_type == np.float64:
        return "double"
    else:
        return "UNKNOWN"



import sys
import re
import inspect
import importlib
from . import bhary


def _target_modules(targets):
    mods = []

    for m in sys.modules:
        for target in targets:
            if m.startswith(target):
                mods.append(m)
    ret = []

    for m in set(mods):
        ret.append(m.replace("numpy.", "numpy_force."))

    return ret


# We will ignore the numpy.core module
IGNORE_MODULES = '[numpy|numpy_force]\.core'

pattern_return_ndarray = re.compile("Return.*ndarray", re.DOTALL)


def modules(targets=["numpy", "numpy_force"], regex_to_exclude=IGNORE_MODULES):
    for m_name in _target_modules(targets):
        if m_name == "numpy":
            # At this point 'numpy' refers to Bohrium and we don't need to bohriumfy Bohrium
            continue

        if re.search(regex_to_exclude, m_name) is not None:
            continue

        try:
            m_obj = importlib.import_module(m_name)
        except:
            continue

        for name, val in inspect.getmembers(m_obj, predicate=inspect.isroutine):
            if not hasattr(val, "_wrapped_fix_biclass"):
                if hasattr(val, "__doc__") and val.__doc__ is not None:
                    if pattern_return_ndarray.search(val.__doc__) is not None:
                        setattr(m_obj, name, bhary.fix_biclass_wrapper(val))
                        # The function should have its original name and not `inner`
                        getattr(m_obj, name).__name__ = name



"""
View Constructors
=================
"""

from __future__ import print_function
import itertools
from .bhary import fix_biclass_wrapper


@fix_biclass_wrapper
def cartesian(x, size):
    dist = []
    for i in range(size):
        start = i
        stop = -size + 1 + i
        if stop == 0: stop = None
        dist.append((start, stop))
    stencil = [x[tuple(s)] for s in [map((lambda se: slice(se[0], se[1])), i)
                                     for i in itertools.product(dist,
                                                                repeat=len(x.shape))]]
    return stencil


@fix_biclass_wrapper
def no_border(x, boarder):
    stencil = [x[tuple(s)] for s in [map((lambda se: slice(se[0], se[1])), i)
                                     for i in itertools.product([(boarder, -boarder)],
                                                                repeat=len(x.shape))]]
    return stencil[0]


@fix_biclass_wrapper
def D2P9(x):
    if len(x.shape) != 2:
        raise Exception('Invalid shape for stencil' + str(len(x)))
    return cartesian(x, 3)


@fix_biclass_wrapper
def D3P27(x):
    if len(x.shape) != 3:
        print(len(x))
        raise Exception('Invalid shape for stencil')
    return cartesian(x, 3)


@fix_biclass_wrapper
def D2P8(x):
    result = D2P9(x)
    result.pop(4)
    return result


@fix_biclass_wrapper
def D3P26(x):
    result = D3P27(x)
    result.pop(13)
    return result


@fix_biclass_wrapper
def grid(x, size):  # 2d for now - should be made nd
    if size != 1:
        raise Exception("Only distance 1 supported")

    data = D2P9(x)
    result = []
    for i in range(9):
        if i % 2 != 0:
            result.append(data[i])

    return result


@fix_biclass_wrapper
def diagonals(x, size):
    if size != 1:
        raise Exception("Only distance 1 supported")

    data = D2P9(x)
    result = []
    for i in range(9):
        if i % 2 == 0:
            result.append(data[i])
    result.pop(2)
    return result



"""
Open Source Computer Vision (OpenCV)
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import bohrium as np
from sys import stderr
from . import ufuncs


def __threshold(name, a, thresh, maxval):
    b = np.array([thresh, maxval], dtype=a.dtype)
    c = np.empty_like(a)
    ufuncs.extmethod(name, c, a, b)
    return c


def threshold(a, thresh=127, maxval=255):
    return __threshold("opencv_threshold_b", a, thresh, maxval)


def threshold_b(a, thresh=127, maxval=255):
    return __threshold("opencv_threshold_b", a, thresh, maxval)


def threshold_bi(a, thresh=127, maxval=255):
    return __threshold("opencv_threshold_bi", a, thresh, maxval)


def threshold_t(a, thresh=127, maxval=255):
    return __threshold("opencv_threshold_t", a, thresh, maxval)


def threshold_tz(a, thresh=127, maxval=255):
    return __threshold("opencv_threshold_tz", a, thresh, maxval)


def threshold_tzi(a, thresh=127, maxval=255):
    return __threshold("opencv_threshold_tzi", a, thresh, maxval)


def connected_components(a, connectivity=8):
    b = np.array([connectivity], dtype=a.dtype)
    c = np.zeros_like(a)
    ufuncs.extmethod("opencv_connected_components", c, a, b)
    return c


def erode(a, b=None, c=None):
    if b is None:
        b = np.array([[0, 1, 0], [1, 1, 1], [0, 1, 0]], dtype=a.dtype)

    if c is None:
        c = np.empty_like(a)

    ufuncs.extmethod("opencv_erode", c, a, b)
    return c


def dilate(a, b=None, c=None):
    if b is None:
        b = np.array([[0, 1, 0], [1, 1, 1], [0, 1, 0]], dtype=a.dtype)

    if c is None:
        c = np.empty_like(a)

    ufuncs.extmethod("opencv_dilate", c, a, b)
    return c



"""
Basic Linear Algebra Subprograms (BLAS)
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Utilize BLAS directly from Python
"""

import bohrium as np
from sys import stderr
from . import ufuncs


def __blas(name, a, b, alpha=1.0, c=None, beta=0.0, shape_matters=True):
    if not b is None:
        if not (a.ndim == 2 and b.ndim == 2):
            stderr.write("[ext] Matrices need to be two-dimensional.\n")
            return None

        if a.shape[1] != b.shape[0] and shape_matters:
            stderr.write(
                "[ext] Wrong shape of matrices: first argument has shape {} and second has shape {}.\n".format(a.shape,
                                                                                                               b.shape))
            return None

        if not b.flags['C_CONTIGUOUS']:
            b = b.copy()
    else:
        b = np.empty(shape=(a.shape[0], a.shape[1]), dtype=a.dtype)

    if not a.flags['C_CONTIGUOUS']:
        a = a.copy()

    if c is None:
        c = np.empty(shape=(a.shape[0], b.shape[1]), dtype=a.dtype)
    elif not c.flags['C_CONTIGUOUS']:
        c = c.copy()

    if alpha != 1.0:
        a = a * alpha

    if beta != 0.0:
        c = c * beta

    ufuncs.extmethod(name, c, a, b)  # modifies 'c'
    return c


# All of A, B, and C are used
def gemm(a, b, alpha=1.0, c=None, beta=0.0):
    """ C := alpha * A * B + beta * C """
    return __blas("blas_gemm", a, b, alpha, c, beta)


def gemmt(a, b, alpha=1.0, c=None, beta=0.0):
    """ C := alpha * A^T * B + beta * C """
    return __blas("blas_gemmt", a, b, alpha, c, beta, shape_matters=False)


def symm(a, b, alpha=1.0, c=None, beta=0.0):
    """ C := alpha * A * B + beta * C """
    """ Notes: A is a symmetric matrix """
    return __blas("blas_symm", a, b, alpha, c, beta)


def hemm(a, b, alpha=1.0, c=None, beta=0.0):
    """ C := alpha * A * B + beta * C """
    """ Notes: A is a hermitian matrix """
    return __blas("blas_hemm", a, b, alpha, c, beta)


def syr2k(a, b, alpha=1.0, c=None, beta=0.0):
    """ C := alpha * A * B**T + alpha * B * A**T + beta * C """
    """ Notes: C is a symmetric matrix """
    return __blas("blas_syr2k", a, b, alpha, c, beta)


def her2k(a, b, alpha=1.0, c=None, beta=0.0):
    """ C := alpha * A * B**H + conjg(alpha) * B * A**H + beta * C """
    """ Notes: C is a hermitian matrix """
    return __blas("blas_her2k", a, b, alpha, c, beta)


# Only A and C are used
def syrk(a, alpha=1.0, c=None, beta=0.0):
    """ C := alpha * A * A**T + beta * C """
    """ Notes: C is a symmetric matrix """
    return __blas("blas_syrk", a, None, alpha, c, beta)


def herk(a, alpha=1.0, c=None, beta=0.0):
    """ C := alpha * A * A**H + beta * C """
    """ Notes: C is a hermitian matrix """
    return __blas("blas_herk", a, None, alpha, c, beta)


# Only A and B are used
def trmm(a, b, alpha=1.0):
    """ B := alpha * A * B """
    """ Notes: A is unit upper triangular matrix """
    __blas("blas_trmm", a, b, alpha)
    return b


def trsm(a, b):
    """ Solves: A * X = B """
    """ Notes: A is unit upper triangular matrix """
    __blas("blas_trsm", a, b)
    return b



import math
import numpy_force as np
from bohrium_api import stack_info
from .. import interop_pyopencl
from .. import interop_pycuda
from .. import array_create
from .bincount_cython import bincount_cython


def bincount(x, weights=None, minlength=None):
    """Count number of occurrences of each value in array of non-negative ints.

    The number of bins (of size 1) is one larger than the largest value in
    `x`. If `minlength` is specified, there will be at least this number
    of bins in the output array (though it will be longer if necessary,
    depending on the contents of `x`).
    Each bin gives the number of occurrences of its index value in `x`.
    If `weights` is specified the input array is weighted by it, i.e. if a
    value ``n`` is found at position ``i``, ``out[n] += weight[i]`` instead
    of ``out[n] += 1``.

    Parameters
    ----------
    x : array_like, 1 dimension, nonnegative ints
        Input array.
    weights : array_like, optional
        Weights, array of the same shape as `x`.
    minlength : int, optional
        A minimum number of bins for the output array.

        .. versionadded:: 1.6.0

    Returns
    -------
    out : ndarray of ints
        The result of binning the input array.
        The length of `out` is equal to ``np.amax(x)+1``.

    Raises
    ------
    ValueError
        If the input is not 1-dimensional, or contains elements with negative
        values, or if `minlength` is non-positive.
    TypeError
        If the type of the input is float or complex.

    See Also
    --------
    histogram, digitize, unique

    Examples
    --------
    >>> np.bincount(np.arange(5))
    array([1, 1, 1, 1, 1])
    >>> np.bincount(np.array([0, 1, 1, 3, 2, 1, 7]))
    array([1, 3, 1, 1, 0, 0, 0, 1])

    >>> x = np.array([0, 1, 1, 3, 2, 1, 7, 23])
    >>> np.bincount(x).size == np.amax(x)+1
    True

    The input array needs to be of integer dtype, otherwise a
    TypeError is raised:

    >>> np.bincount(np.arange(5, dtype=np.float))
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
    TypeError: array cannot be safely cast to required type

    A possible use of ``bincount`` is to perform sums over
    variable-size chunks of an array, using the ``weights`` keyword.

    >>> w = np.array([0.3, 0.5, 0.2, 0.7, 1., -0.6]) # weights
    >>> x = np.array([0, 1, 1, 2, 2, 2])
    >>> np.bincount(x,  weights=w)
    array([ 0.3,  0.7,  1.1])

    """

    # Let's find the backend to handle the bincount
    x = array_create.array(x)
    assert(np.issubdtype(x.dtype.type, np.integer))
    assert(np.issubdtype(x.dtype.type, np.integer))

    if stack_info.is_proxy_in_stack():  # Cannot directly access array data through a proxy
        return np.bincount(x.copy2numpy(), weights=weights, minlength=minlength)

    try:
        if weights is not None:
            raise NotImplementedError("OpenCL doesn't support the `weights` argument")
        return bincount_pyopencl(x, minlength=minlength)
    except NotImplementedError:
        try:
            if weights is not None:
                raise NotImplementedError("CUDA doesn't support the `weights` argument")
            return bincount_pycuda(x, minlength=minlength)
        except NotImplementedError:
            try:
                return bincount_cython(x, weights=weights, minlength=minlength)
            except NotImplementedError:
                return np.bincount(x.copy2numpy(), weights=weights, minlength=minlength)


def bincount_pyopencl(x, minlength=None):
    """PyOpenCL implementation of `bincount()`"""

    if not interop_pyopencl.available():
        raise NotImplementedError("OpenCL not available")

    import pyopencl as cl
    ctx = interop_pyopencl.get_context()
    queue = cl.CommandQueue(ctx)

    x_max = int(x.max())
    if x_max < 0:
        raise RuntimeError("bincount(): first argument must be a 1 dimensional, non-negative int array")
    if x_max > np.iinfo(np.uint32).max:
        raise NotImplementedError("OpenCL: the elements in the first argument must fit in a 32bit integer")
    if minlength is not None:
        x_max = max(x_max, minlength)

    # TODO: handle large max element by running multiple bincount() on a range
    if x_max >= interop_pyopencl.max_local_memory(queue.device) // x.itemsize:
        raise NotImplementedError("OpenCL: max element is too large for the GPU")

    # Let's create the output array and retrieve the in-/output OpenCL buffers
    # NB: we always return uint32 array
    ret = array_create.empty((x_max+1, ), dtype=np.uint32)
    x_buf = interop_pyopencl.get_buffer(x)
    ret_buf = interop_pyopencl.get_buffer(ret)

    # OpenCL kernel is based on the book "OpenCL Programming Guide" by Aaftab Munshi at al.
    source = """
    kernel void histogram_partial(
        global DTYPE *input,
        global uint *partial_histo,
        uint input_size
    ){
        int local_size = (int)get_local_size(0);
        int group_indx = get_group_id(0) * HISTO_SIZE;
        int gid = get_global_id(0);
        int tid = get_local_id(0);

        local uint tmp_histogram[HISTO_SIZE];

        int j = HISTO_SIZE;
        int indx = 0;

        // clear the local buffer that will generate the partial histogram
        do {
            if (tid < j)
                tmp_histogram[indx+tid] = 0;
            j -= local_size;
            indx += local_size;
        } while (j > 0);

        barrier(CLK_LOCAL_MEM_FENCE);

        if (gid < input_size) {
            atomic_inc(&tmp_histogram[input[gid]]);
        }

        barrier(CLK_LOCAL_MEM_FENCE);

        // copy the partial histogram to appropriate location in
        // histogram given by group_indx
        if (local_size >= HISTO_SIZE){
            if (tid < HISTO_SIZE)
                partial_histo[group_indx + tid] = tmp_histogram[tid];
        }else{
            j = HISTO_SIZE;
            indx = 0;
            do {
                if (tid < j)
                    partial_histo[group_indx + indx + tid] = tmp_histogram[indx + tid];

                j -= local_size;
                indx += local_size;
            } while (j > 0);
        }
    }

    kernel void histogram_sum_partial_results(
        global uint *partial_histogram,
        int num_groups,
        global uint *histogram
    ){
        int gid = (int)get_global_id(0);
        int group_indx;
        int n = num_groups;
        local uint tmp_histogram[HISTO_SIZE];

        tmp_histogram[gid] = partial_histogram[gid];
        group_indx = HISTO_SIZE;
        while (--n > 0) {
            tmp_histogram[gid] += partial_histogram[group_indx + gid];
            group_indx += HISTO_SIZE;
        }
        histogram[gid] = tmp_histogram[gid];
    }
    """
    source = source.replace("HISTO_SIZE", "%d" % ret.shape[0])
    source = source.replace("DTYPE", interop_pyopencl.type_np2opencl_str(x.dtype))
    prg = cl.Program(ctx, source).build()

    # Calculate sizes for the kernel execution
    local_size = interop_pyopencl.kernel_info(prg.histogram_partial, queue)[0]  # Max work-group size
    num_groups = int(math.ceil(x.shape[0] / float(local_size)))
    global_size = local_size * num_groups

    # First we compute the partial histograms
    partial_res_g = cl.Buffer(ctx, cl.mem_flags.WRITE_ONLY, num_groups * ret.nbytes)
    prg.histogram_partial(queue, (global_size,), (local_size,), x_buf, partial_res_g, np.uint32(x.shape[0]))

    # Then we sum the partial histograms into the final histogram
    prg.histogram_sum_partial_results(queue, ret.shape, None, partial_res_g, np.uint32(num_groups), ret_buf)
    return ret


def bincount_pycuda(x, minlength=None):
    """PyCUDA implementation of `bincount()`"""

    if not interop_pycuda.available():
        raise NotImplementedError("CUDA not available")

    import pycuda
    from pycuda.compiler import SourceModule

    interop_pycuda.init()

    x_max = int(x.max())
    if x_max < 0:
        raise RuntimeError("bincount(): first argument must be a 1 dimensional, non-negative int array")
    if x_max > np.iinfo(np.uint32).max:
        raise NotImplementedError("CUDA: the elements in the first argument must fit in a 32bit integer")
    if minlength is not None:
        x_max = max(x_max, minlength)

    # TODO: handle large max element by running multiple bincount() on a range
    if x_max >= interop_pycuda.max_local_memory() // x.itemsize:
        raise NotImplementedError("CUDA: max element is too large for the GPU")

    # Let's create the output array and retrieve the in-/output OpenCL buffers
    # NB: we always return uint32 array
    ret = array_create.ones((x_max+1, ), dtype=np.uint32)
    x_buf = interop_pycuda.get_gpuarray(x)
    ret_buf = interop_pycuda.get_gpuarray(ret)

    # CUDA kernel is based on the book "OpenCL Programming Guide" by Aaftab Munshi at al.
    source = """
    __global__ void histogram_partial(
        DTYPE *input,
        uint *partial_histo,
        uint input_size
    ){
        int local_size = blockDim.x;
        int group_indx = blockIdx.x * HISTO_SIZE;
        int gid = (blockIdx.x * blockDim.x + threadIdx.x);
        int tid = threadIdx.x;

        __shared__ uint tmp_histogram[HISTO_SIZE];

        int j = HISTO_SIZE;
        int indx = 0;

        // clear the local buffer that will generate the partial histogram
        do {
            if (tid < j)
                tmp_histogram[indx+tid] = 0;
            j -= local_size;
            indx += local_size;
        } while (j > 0);

        __syncthreads();

        if (gid < input_size) {
            atomicAdd(&tmp_histogram[input[gid]], 1);
        }

        __syncthreads();

        // copy the partial histogram to appropriate location in
        // histogram given by group_indx
        if (local_size >= HISTO_SIZE){
            if (tid < HISTO_SIZE)
                partial_histo[group_indx + tid] = tmp_histogram[tid];
        }else{
            j = HISTO_SIZE;
            indx = 0;
            do {
                if (tid < j)
                    partial_histo[group_indx + indx + tid] = tmp_histogram[indx + tid];

                j -= local_size;
                indx += local_size;
            } while (j > 0);
        }
    }

    __global__ void histogram_sum_partial_results(
        uint *partial_histogram,
        int num_groups,
        uint *histogram
    ){
        int gid = (blockIdx.x * blockDim.x + threadIdx.x);
        int group_indx;
        int n = num_groups;
        __shared__ uint tmp_histogram[HISTO_SIZE];

        tmp_histogram[gid] = partial_histogram[gid];
        group_indx = HISTO_SIZE;
        while (--n > 0) {
            tmp_histogram[gid] += partial_histogram[group_indx + gid];
            group_indx += HISTO_SIZE;
        }
        histogram[gid] = tmp_histogram[gid];
    }
    """
    source = source.replace("HISTO_SIZE", "%d" % ret.shape[0])
    source = source.replace("DTYPE", interop_pycuda.type_np2cuda_str(x.dtype))
    prg = SourceModule(source)

    # Calculate sizes for the kernel execution
    kernel = prg.get_function("histogram_partial")
    local_size = kernel.get_attribute(pycuda.driver.function_attribute.MAX_THREADS_PER_BLOCK)  # Max work-group size
    num_groups = int(math.ceil(x.shape[0] / float(local_size)))
    global_size = local_size * num_groups

    # First we compute the partial histograms
    partial_res_g = pycuda.driver.mem_alloc(num_groups * ret.nbytes)
    kernel(x_buf, partial_res_g, np.uint32(x.shape[0]), block=(local_size, 1, 1), grid=(num_groups, 1))

    # Then we sum the partial histograms into the final histogram
    kernel = prg.get_function("histogram_sum_partial_results")
    kernel(partial_res_g, np.uint32(num_groups), ret_buf, block=(1, 1, 1), grid=(ret.shape[0], 1))
    return ret  





"""
Visualization
~~~~~~~~~~~~~

Common functions for visualization.
"""
import bohrium as np
from . import ufuncs, array_create
from bohrium import _bh


def compressed_copy(ary, param):
    a_min = ary.min()
    a_range = ary.max() - ary.min() + 1
    # Normalize `ary` into uint8
    a = (ary - a_min) * 256 / a_range
    assert (a.min() >= 0)
    assert (a.max() < 256)
    a = np.array(a, dtype=np.uint8)
    # Copy `a` using `param`
    a = _bh.mem_copy(a, param=param)
    # un-normalize and convert back to the original dtype of `ary`
    a = array_create.array(a, dtype=ary.dtype)
    return (a * a_range + a_min) / 256.0


def plot_surface(ary, mode, colormap, lowerbound, upperbound, param=None):
    mode = mode.lower()

    ranks = [2, 3]
    modes = ["2d", "3d"]
    types = [np.float32]

    if not (ary.ndim == 2 or ary.ndim == 3):
        raise ValueError("Unsupported array-rank, must be one of %s" % ranks)
    if mode not in modes:
        raise ValueError("Unsupported mode, must be one of %s" % modes)

    if ary.dtype not in types:
        ary = array_create.array(ary, bohrium=True, dtype=np.float32)

    if mode == "2d":
        flat = True
        cube = False
    elif mode == "3d":
        if ary.ndim == 2:
            flat = False
            cube = False
        else:
            flat = False
            cube = True
    else:
        raise ValueError("Unsupported mode '%s' " % mode)

    args = array_create.array([  # Construct arguments
        np.float32(colormap),
        np.float32(flat),
        np.float32(cube),
        np.float32(lowerbound),
        np.float32(upperbound)
    ],
        bohrium=True,
        dtype=np.float32
    )

    if param is not None:
        ary = compressed_copy(ary, param)

    ufuncs.extmethod("visualizer", ary, args, ary)  # Send to extension



"""
Signal Processing
~~~~~~~~~~~~~~~~~

Common signal processing functions, which often handle multiple dimension

"""
import numpy_force as numpy
from . import array_create
from . import bhary
from . import ufuncs
from . import linalg
from . import summations
from . import numpy_backport
from . import _bh


# 1d
# ---------------------------------------------------------------------------------
def _correlate_and_convolve_body(vector, filter, d, mode):
    """ The body of correlate() and convolve() are identical"""

    # Anything to do?
    if vector.size <= 0:
        return vector.copy()

    # Complex correlate includes a conjugation
    if numpy.iscomplexobj(filter):
        filter = numpy.conj(filter)

    dtype = numpy.result_type(vector, filter)
    rows = int(ufuncs.ceil((vector.size + 2 * filter.size - 2) / float(filter.size)))
    padded = array_create.empty([rows * filter.size], dtype=dtype)
    padded[0:filter.size - 1] = 0
    padded[filter.size - 1:vector.size + filter.size - 1] = vector
    padded[vector.size + filter.size - 1:] = 0
    s = numpy_backport.as_strided(padded, shape=(padded.shape[0] - filter.size + 1, filter.size),
                                  strides=(padded.strides[0], padded.strides[0]))
    result = linalg.dot(s, filter)
    if mode == 'same':
        return result[d:vector.size + d]
    elif mode == 'full':
        return result[0:vector.size + filter.size - 1]
    elif mode == 'valid':
        return result[filter.size - 1:vector.size]
    else:
        raise ValueError("correlate1d: invalid mode '%s'" % mode)


def correlate1d(a, v, mode='valid'):
    assert a.ndim == 1
    assert v.ndim == 1

    # Let's make sure that we are working on Bohrium arrays
    if not bhary.check(a) and not bhary.check(v):
        return numpy.correlate(a, v, mode)
    else:
        a = array_create.array(a)
        v = array_create.array(v)

    # We might have to swap `a` and `v` such that `vector` is always largest
    if v.shape[0] > a.shape[0]:
        vector = v[::-1]
        filter = a[::-1]
        d = int(filter.size / 2)
    else:
        vector = a
        filter = v
        d = int((filter.size - 1) / 2)
    return _correlate_and_convolve_body(vector, filter, d, mode)


def convolve1d(a, v, mode='full'):
    assert a.ndim == 1
    assert v.ndim == 1

    # Let's make sure that we are working on Bohrium arrays
    if not bhary.check(a) and not bhary.check(v):
        return numpy.convolve(a, v, mode)
    else:
        a = array_create.array(a)
        v = array_create.array(v)

    # We might have to swap `a` and `v` such that `vector` is always largest
    if v.shape[0] > a.shape[0]:
        vector = v
        filter = a[::-1]
    else:
        vector = a
        filter = v[::-1]
    d = int((filter.size - 1) / 2)
    return _correlate_and_convolve_body(vector, filter, d, mode)


# Nd
# ---------------------------------------------------------------------------------
def _findIndices(ArrSize, FilterSize):
    N = FilterSize.shape[0]
    n = int(FilterSize.prod())
    CumSizeArr = numpy.ones([N], dtype=numpy.int32)
    CumSizeArr[1:N] = ArrSize[0:N - 1].cumprod()
    CumSize = numpy.ones([N], dtype=numpy.int32)
    CumSize[1:N] = FilterSize[0:N - 1].cumprod()

    vals = numpy.empty((n, N), dtype=numpy.int32)
    for i in range(N):
        vals[:, i] = numpy.linspace(0, n - 1, n)

    vals = vals // CumSize
    vals = vals % FilterSize
    CurrPos = summations.sum(vals * CumSizeArr, axis=1)

    return CurrPos.astype(numpy.int32)


def _addZerosNd(Array, FilterSize, dtype):
    # Introduces zero padding for Column major flattening
    PaddedSize = numpy.array(Array.shape, dtype=numpy.int32)
    N = FilterSize.shape[0]
    PaddedSize[0:N] += FilterSize - 1
    cut = '['
    for i in range(PaddedSize.shape[0]):
        if i < N:
            minpos = int(FilterSize[i] / 2)
            maxpos = Array.shape[i] + int(FilterSize[i] / 2)
        else:
            minpos = 0
            maxpos = Array.shape[i]
        cut += str(minpos) + ':' + str(maxpos) + ','
    cut = cut[:-1] + ']'
    Padded = array_create.zeros(PaddedSize, dtype=dtype, bohrium=bhary.check(Array))
    exec ('Padded' + cut + '=Array')
    return Padded


def _findSame(Array, FilterSize):
    # Numpy convention. Returns view of the same size as the largest input
    N = FilterSize.shape[0]
    cut = '['
    for i in range(len(Array.shape)):
        if i < N:
            minpos = (FilterSize[i] - 1) // 2
            maxpos = Array.shape[i] - (FilterSize[i]) // 2
        else:
            minpos = 0
            maxpos = Array.shape[i]
        cut += str(minpos) + ':' + str(maxpos) + ','
    cut = cut[:-1] + ']'
    res = eval('Array' + cut)
    return res


def _findValid(Array, FilterSize):
    # Cuts the result down to only totally overlapping views
    N = FilterSize.shape[0]
    cut = '['
    for i in range(len(Array.shape)):
        if i < N:
            minpos = FilterSize[i] - 1
            maxpos = Array.shape[i] - FilterSize[i] + 1
        else:
            minpos = 0
            maxpos = Array.shape[i]
        cut += str(minpos) + ':' + str(maxpos) + ','
    cut = cut[:-1] + ']'
    res = eval('Array' + cut)
    return res


def _invert_ary(a):
    """Reverse all elements in each axis"""

    def flip(m, axis):
        """Copy of `numpy.flip()`, which were introduced in NumPy v1.12"""
        indexer = [slice(None)] * m.ndim
        try:
            indexer[axis] = slice(None, None, -1)
        except IndexError:
            raise ValueError("axis=%i is invalid for the %i-dimensional input array"
                             % (axis, m.ndim))
        return m[tuple(indexer)]

    # Flip all axises
    for i in range(len(a.shape)):
        a = flip(a, axis=i)
    return a


def _correlate_kernel(Array, Filter, mode):
    # Anything to do?
    if Array.size <= 0:
        return Array.copy()

    # Complex correlate includes a conjugation
    if numpy.iscomplexobj(Filter):
        Filter = numpy.conj(Filter)

    # Get sizes as arrays for easier manipulation
    ArrSize = numpy.array(Array.shape, dtype=numpy.int32)
    FilterSize = numpy.array(Filter.shape, dtype=numpy.int32)

    # Check that mode='valid' is allowed given the array sizes
    if mode == 'valid':
        diffSize = ArrSize[:FilterSize.size] - FilterSize
        nSmaller = summations.sum(diffSize < 0)
        if nSmaller > 0:
            raise ValueError(
                "correlateNd: For 'valid' mode, one must be at least as large as the other in every dimension")

    # Use numpy convention for result dype
    dtype = numpy.result_type(Array, Filter)

    # Add zeros along relevant dimensions
    Padded = _addZerosNd(Array, FilterSize, dtype)
    PaddedSize = numpy.array(Padded.shape, dtype=numpy.int32)

    # Get positions of first view
    IndiVec = _findIndices(PaddedSize, FilterSize)
    CenterPos = tuple((FilterSize - 1) // 2)
    IndiMat = IndiVec.reshape(FilterSize, order='F')
    nPre = IndiMat[CenterPos]  # Required zeros before Array for correct alignment
    nPost = IndiVec[Filter.size - 1] - nPre
    n = Padded.size
    nTot = n + nPre + nPost  # Total size after pre/post padding
    V = array_create.empty([nTot], dtype=dtype, bohrium=bhary.check(Array))
    V[nPre:n + nPre] = Padded.flatten(order='F')
    V[:nPre] = 0
    V[n + nPre:] = 0
    A = Filter.flatten(order='F')

    # Actual correlation calculation
    Correlated = V[IndiVec[0]:n + IndiVec[0]] * A[0]
    for i in range(1, Filter.size):
        Correlated += V[IndiVec[i]:n + IndiVec[i]] * A[i]
        # TODO: we need this flush because of very slow fusion
        if bhary.check(V):
            _bh.flush()

    Full = Correlated.reshape(PaddedSize, order='F')
    if mode == 'full':
        return Full
    elif mode == 'same':
        return _findSame(Full, FilterSize)
    elif mode == 'valid':
        return _findValid(Full, FilterSize)
    else:
        raise ValueError("correlateNd: invalid mode '%s'" % mode)


def convolve(a, v, mode='full'):
    if (a.size > v.size) or (mode == 'same'):
        Array = a
        Filter = _invert_ary(v)
    else:
        Array = v
        Filter = _invert_ary(a)
    return _correlate_kernel(Array, Filter, mode)


def correlate(a, v, mode='valid'):
    if (a.size > v.size) or (mode == 'same'):
        Array = a
        Filter = v
    else:
        Array = _invert_ary(v)
        Filter = _invert_ary(a)
    return _correlate_kernel(Array, Filter, mode)



"""
Array manipulation routines
===========================
"""
from copy import deepcopy
from . import array_create
import numpy_force as numpy
from . import bhary
from . import _util
from .bhary import fix_biclass_wrapper
from . import numpy_backport
from . import loop


@fix_biclass_wrapper
def flatten(ary, order='C', always_copy=True):
    """
    Return a copy of the array collapsed into one dimension.

    Parameters
    ----------
    ary : array_like
        Array from which to retrieve the flattened data from.
    order : {'C', 'F', 'A', 'K'}, optional
        'C' means to flatten in row-major (C-style) order.
        'F' means to flatten in column-major (Fortran-
        style) order. 'A' means to flatten in column-major
        order if `a` is Fortran *contiguous* in memory,
        row-major order otherwise. 'K' means to flatten
        `a` in the order the elements occur in memory.
        The default is 'C'.
    always_copy : boolean
        When False, a copy is only made when necessary

    Returns
    -------
    y : ndarray
        A copy of the input array, flattened to one dimension.

    Notes
    -----
    The order of the data in memory is always row-major (C-style).

    Examples
    --------
    >>> a = np.array([[1,2], [3,4]])
    >>> np.flatten(a)
    array([1, 2, 3, 4])
    """

    if order == 'F' or (order == 'A' and not ary.flags['F_CONTIGUOUS']):
        ary = numpy.transpose(ary)

    ret = ary.reshape(_util.totalsize(ary))
    if always_copy:
        return ret.copy()
    else:
        return ret


@fix_biclass_wrapper
def diagonal(ary, offset=0, axis1=0, axis2=1):
    """
    Return specified diagonals.

    If `a` is 2-D, returns the diagonal of `a` with the given offset,
    i.e., the collection of elements of the form ``a[i, i+offset]``.
    If `a` has more than two dimensions, then the axes specified by
    `axis1` and `axis2` are used to determine the 2-D sub-array whose
    diagonal is returned. The shape of the resulting array can be
    determined by removing `axis1` and `axis2` and appending an index
    to the right equal to the size of the resulting diagonals.

    Parameters
    ----------
    ary : array_like
        Array from which the diagonals are taken.
    offset : int, optional
        Offset of the diagonal from the main diagonal.  Can be positive or
        negative.  Defaults to main diagonal (0).
    axis1 : int, optional
        Axis to be used as the first axis of the 2-D sub-arrays from which
        the diagonals should be taken. Defaults to first axis (0).
    axis2 : int, optional
        Axis to be used as the second axis of the 2-D sub-arrays from which
        the diagonals should be taken. Defaults to second axis (1).

    Returns
    -------
    array_of_diagonals : ndarray
        If `a` is 2-D, a 1-D array containing the diagonal is returned.
        If the dimension of `a` is larger, then an array of diagonals is
        returned, "packed" from left-most dimension to right-most (e.g.,
        if `a` is 3-D, then the diagonals are "packed" along rows).

    Raises
    ------
    ValueError
        If the dimension of `a` is less than 2.

    See Also
    --------
    diag : MATLAB work-a-like for 1-D and 2-D arrays.
    diagflat : Create diagonal arrays.
    trace : Sum along diagonals.

    Examples
    --------
    >>> a = np.arange(4).reshape(2,2); a
    array([[0, 1],
           [2, 3]])
    >>> a.diagonal()
    array([0, 3])
    >>> a.diagonal(1)
    array([1])

    A 3-D example:

    >>> a = np.arange(8).reshape(2,2,2); a
    array([[[0, 1],
            [2, 3]],

           [[4, 5],
            [6, 7]]])
    >>> a.diagonal()
    array([[0, 6],
           [1, 7]])
    """
    if axis1 == axis2:
        raise Exception("axis1 and axis2 cannot be the same\n")
    if ary.ndim < 2:
        raise Exception("diagonal requires an array of at least two dimensions\n")

    # Get all axes except the two which has the diagonal we seek;
    # these are added later
    min_axis, max_axis = sorted([axis1, axis2])
    tr = list(range(ary.ndim))
    del tr[max_axis]
    del tr[min_axis]

    # Positive offset means upper diagonals, negative is lower, so we switch
    # the axes around if negative
    if offset >= 0:
        ary = ary.transpose(tr + [axis1, axis2])
    else:
        ary = ary.transpose(tr + [axis2, axis1])
        offset = -offset

    # Calculate how many elements will be in the diagonal
    diag_size = max(0, min(ary.shape[-2], ary.shape[-1] - offset))
    ret_shape = ary.shape[:-2] + (diag_size,)

    # Return empty array if the diagonal has zero elements
    if diag_size == 0:
        return array_create.empty(ret_shape, dtype=ary.dtype, bohrium=bhary.check(ary))

    ary = ary[..., :diag_size, offset:(offset + diag_size)]

    ret_strides = ary.strides[:-2] + (ary.strides[-1] + ary.strides[-2],)
    return numpy_backport.as_strided(ary, shape=ret_shape, strides=ret_strides)


@fix_biclass_wrapper
def diagflat(d, k=0):
    """
    Create a two-dimensional array with the flattened input as a diagonal.

    Parameters
    ----------
    d : array_like
        Input data, which is flattened and set as the `k`-th
        diagonal of the output.
    k : int, optional
        Diagonal to set; 0, the default, corresponds to the "main" diagonal,
        a positive (negative) `k` giving the number of the diagonal above
        (below) the main.

    Returns
    -------
    out : ndarray
        The 2-D output array.

    See Also
    --------
    diag : MATLAB work-alike for 1-D and 2-D arrays.
    diagonal : Return specified diagonals.
    trace : Sum along diagonals.

    Examples
    --------
    >>> np.diagflat([[1,2], [3,4]])
    array([[1, 0, 0, 0],
           [0, 2, 0, 0],
           [0, 0, 3, 0],
           [0, 0, 0, 4]])

    >>> np.diagflat([1,2], 1)
    array([[0, 1, 0],
           [0, 0, 2],
           [0, 0, 0]])

    """
    d = flatten(d)
    size = d.size + abs(k)
    A = array_create.zeros((size, size), dtype=d.dtype, bohrium=bhary.check(d))
    Ad = diagonal(A, offset=k)
    Ad[...] = d
    return A


@fix_biclass_wrapper
def diag(v, k=0):
    """
    Extract a diagonal or construct a diagonal array.

    Parameters
    ----------
    v : array_like
        If `v` is a 2-D array, return a copy of its `k`-th diagonal.
        If `v` is a 1-D array, return a 2-D array with `v` on the `k`-th
        diagonal.
    k : int, optional
        Diagonal in question. The default is 0. Use `k>0` for diagonals
        above the main diagonal, and `k<0` for diagonals below the main
        diagonal.

    Returns
    -------
    out : ndarray
        The extracted diagonal or constructed diagonal array.

    See Also
    --------
    diagonal : Return specified diagonals.
    diagflat : Create a 2-D array with the flattened input as a diagonal.
    trace : Sum along diagonals.
    triu : Upper triangle of an array.
    tril : Lower triange of an array.

    Examples
    --------
    >>> x = np.arange(9).reshape((3,3))
    >>> x
    array([[0, 1, 2],
           [3, 4, 5],
           [6, 7, 8]])

    >>> np.diag(x)
    array([0, 4, 8])
    >>> np.diag(x, k=1)
    array([1, 5])
    >>> np.diag(x, k=-1)
    array([3, 7])

    >>> np.diag(np.diag(x))
    array([[0, 0, 0],
           [0, 4, 0],
           [0, 0, 8]])
    """

    if v.ndim == 1:
        return diagflat(v, k)
    elif v.ndim == 2:
        return diagonal(v, k)
    else:
        raise ValueError("Input must be 1- or 2-d.")


@fix_biclass_wrapper
def reshape(a, *newshape, **kwargs):
    """
    Gives a new shape to an array without changing its data.

    Parameters
    ----------
    a : array_like
        Array to be reshaped.
    newshape : int or tuple of ints
        The new shape should be compatible with the original shape. If
        an integer, then the result will be a 1-D array of that length.
        One shape dimension can be -1. In this case, the value is inferred
        from the length of the array and remaining dimensions.
    order : {`C`, `F`, `A`}, optional
        Read the elements of a using this index order, and place the elements
        into the reshaped array using this index order.
        `C` means to read / write the elements using C-like index order,
        with the last axis index changing fastest, back to the first axis
        index changing slowest. `F` means to read / write the elements using
        Fortran-like index order, with the first index changing fastest,
        and the last index changing slowest. Note that the `C` and `F` options
        take no account of the memory layout of the underlying array,
        and only refer to the order of indexing. `A` means to read / write
        the elements in Fortran-like index order if a is Fortran contiguous
        in memory, C-like order otherwise.

    Returns
    -------
    reshaped_array : ndarray
        This will be a new view object if possible; otherwise, it will
        be a copy.  Note there is no guarantee of the *memory layout* (C- or
        Fortran- contiguous) of the returned array.

    See Also
    --------
    ndarray.reshape : Equivalent method.

    Notes
    -----
    It is not always possible to change the shape of an array without
    copying the data. If you want an error to be raise if the data is copied,
    you should assign the new shape to the shape attribute of the array::

     >>> a = np.zeros((10, 2))
     # A transpose make the array non-contiguous
     >>> b = a.T
     # Taking a view makes it possible to modify the shape without modifying the
     # initial object.
     >>> c = b.view()
     >>> c.shape = (20)
     AttributeError: incompatible shape for a non-contiguous array

    The `order` keyword gives the index ordering both for *fetching* the values
    from `a`, and then *placing* the values into the output array.  For example,
    let's say you have an array:

    >>> a = np.arange(6).reshape((3, 2))
    >>> a
    array([[0, 1],
           [2, 3],
           [4, 5]])

    You can think of reshaping as first raveling the array (using the given
    index order), then inserting the elements from the raveled array into the
    new array using the same kind of index ordering as was used for the
    raveling.

    >>> np.reshape(a, (2, 3)) # C-like index ordering
    array([[0, 1, 2],
           [3, 4, 5]])
    >>> np.reshape(np.ravel(a), (2, 3)) # equivalent to C ravel then C reshape
    array([[0, 1, 2],
           [3, 4, 5]])
    >>> np.reshape(a, (2, 3), order='F') # Fortran-like index ordering
    array([[0, 4, 3],
           [2, 1, 5]])
    >>> np.reshape(np.ravel(a, order='F'), (2, 3), order='F')
    array([[0, 4, 3],
           [2, 1, 5]])

    Examples
    --------
    >>> a = np.array([[1,2,3], [4,5,6]])
    >>> np.reshape(a, 6)
    array([1, 2, 3, 4, 5, 6])
    >>> np.reshape(a, 6, order='F')
    array([1, 4, 2, 5, 3, 6])

    >>> np.reshape(a, (3,-1))       # the unspecified value is inferred to be 2
    array([[1, 2],
           [3, 4],
           [5, 6]])
    """
    # Let's make sure that newshape is a flat sequence
    if len(newshape) == 1:
        # The item is a sequence
        if hasattr(newshape[0], "__getitem__"):
            newshape = newshape[0]

    if not a.flags['C_CONTIGUOUS']:
        t = array_create.empty_like(a)
        t[...] = a
        a = t

    return numpy.ndarray.reshape(a, newshape, **kwargs)


@fix_biclass_wrapper
def trace(ary, offset=0, axis1=0, axis2=1, dtype=None):
    """
    Return the sum along diagonals of the array.

    If `a` is 2-D, the sum along its diagonal with the given offset
    is returned, i.e., the sum of elements ``a[i,i+offset]`` for all i.

    If `a` has more than two dimensions, then the axes specified by axis1 and
    axis2 are used to determine the 2-D sub-arrays whose traces are returned.
    The shape of the resulting array is the same as that of `a` with `axis1`
    and `axis2` removed.

    Parameters
    ----------
    ary : array_like
        Input array, from which the diagonals are taken.
    offset : int, optional
        Offset of the diagonal from the main diagonal. Can be both positive
        and negative. Defaults to 0.
    axis1, axis2 : int, optional
        Axes to be used as the first and second axis of the 2-D sub-arrays
        from which the diagonals should be taken. Defaults are the first two
        axes of `a`.
    dtype : dtype, optional
        Determines the data-type of the returned array and of the accumulator
        where the elements are summed. If dtype has the value None and `a` is
        of integer type of precision less than the default integer
        precision, then the default integer precision is used. Otherwise,
        the precision is the same as that of `a`.
    out : ndarray, optional
        Array into which the output is placed. Its type is preserved and
        it must be of the right shape to hold the output.

    Returns
    -------
    sum_along_diagonals : ndarray
        If `a` is 2-D, the sum along the diagonal is returned.  If `a` has
        larger dimensions, then an array of sums along diagonals is returned.

    See Also
    --------
    diag, diagonal, diagflat

    Examples
    --------
    >>> np.trace(np.eye(3))
    3.0
    >>> a = np.arange(8).reshape((2,2,2))
    >>> np.trace(a)
    array([6, 8])

    >>> a = np.arange(24).reshape((2,2,2,3))
    >>> np.trace(a).shape
    (2, 3)

    """
    D = diagonal(ary, offset=offset, axis1=axis1, axis2=axis2)

    if dtype:
        D = D.astype(dtype)

    return D.sum(axis=-1)


@fix_biclass_wrapper
def broadcast_arrays(*args):
    """
    Broadcast any number of arrays against each other.

    .. note:: This function is very similar to NumPy's  `broadcast_arrays()`

    Parameters
    ----------
    `array_list` : array_likes
        The arrays to broadcast.

    Returns
    -------
    broadcasted : list of arrays
        These arrays are views on the original arrays or the untouched originals.
        They are typically not contiguous.  Furthermore, more than one element of a
        broadcasted array may refer to a single memory location.  If you
        need to write to the arrays, make copies first.
    shape : tuple
        The shape the arrays are broadcasted to
    """
    try:
        if len(args) == 0:
            return ([], [])

        if len(args) == 1:
            if numpy.isscalar(args[0]):  # It is possible that `args[0]` is a scalar
                shape = (1,)
            else:
                shape = args[0].shape
            return (args, shape)

        # Common case where nothing needs to be broadcasted.
        bcast = numpy.broadcast(*args)
        if all(array.shape == bcast.shape for array in args if not numpy.isscalar(array)):
            return (args, bcast.shape)

        ret = []
        # We use NumPy's broadcast_arrays() to broadcast the views.
        # Notice that the 'subok' argument is first introduced in version 10 of NumPy
        try:
            bargs = numpy.broadcast_arrays(*args, subok=True)
        except TypeError as err:
            if "subok" in err.message:
                bargs = numpy.broadcast_arrays(*args)
            else:
                raise

        # The broadcasted view inherits dynamic changes if there are any.
        # Used for broadcasting dynamic views within a do_while loop
        bcast_array = args[0]
        bcast_dvi = bcast_array.bhc_dynamic_view_info
        for a, b in zip(args, bargs):
            # If the broadcast view changes shape between iterations,
            # force the same change to the views being broadcasted.
            # Used in regard to iterators within do_while loops.
            a_dvi = a.bhc_dynamic_view_info

            # If array that is broadcasted has dynamic changes, the
            # broadcasted array must inherit these
            if a_dvi:
                b_dvi = deepcopy(a_dvi)
            else:
                b_dvi = loop.DynamicViewInfo({}, a.shape, a.strides)

            # If the array that is broadcasted from has changes in shape
            # must these changes also be inherited by the broadcasted array
            if bcast_dvi:
                # If the view contains a slide in a broadcasted dimension,
                # the slide must be inherited
                for dim in bcast_dvi.dims_with_changes():
                    # If the array, which is broadcasted from, does not have
                    # dynamic changes, there are no changes to add
                    if bcast_dvi.dim_shape_change(dim) == 0:
                        continue

                    # The array, which is broadcasted from, has dynamic changes
                    # while the broadcasted array does not. Add the changes to the
                    # broadcasted array
                    elif b_dvi.dim_shape_change(dim) == 0:
                        for (_, shape_change, step_delay, shape, stride) in bcast_dvi.changes_in_dim(dim):
                            # No reason to add a change of 0 in the dimension
                            if shape_change == 0:
                                continue
                            b_dvi.add_dynamic_change(dim, 0, shape_change, step_delay, shape, stride)

                    # Both array, which is broadcasted from, and the broadcasted array has
                    # dynamic changes. Make sure they are the same. If not the change cannot
                    # be guessed, which results in an error.
                    elif b_dvi.dim_shape_change(dim) != 0 and \
                            b_dvi.dim_shape_change(dim) != bcast_dvi.dim_shape_change(dim):
                        raise loop.IteratorIllegalBroadcast(
                            dim, a.shape, a_dvi.dim_shape_change(dim),
                            bcast_array.shape, bcast_dvi.dim_shape_change(dim))

            # Add the dynamic changes, if any
            if b_dvi.has_changes():
                b.bhc_dynamic_view_info = b_dvi

            # Append the broadcasted array
            ret.append(b)

    except ValueError as msg:
        if str(msg).find("shape mismatch: objects cannot be broadcast to a single shape") != -1:
            shapes = [arg.shape for arg in args]
            raise ValueError("shape mismatch: objects cannot be broadcasted to a single shape: %s" % shapes)
        raise
    return (ret, bcast.shape)


@fix_biclass_wrapper
def fill(a, value):
    """
    a.fill(value)

    Fill the array with a scalar value.

    Parameters
    ----------
        a : array_like
        Array to fill
    value : scalar
        All elements of `a` will be assigned this value.

    Examples
    --------
    >>> a = np.array([1, 2])
    >>> a.fill(0)
    >>> a
    array([0, 0])
    >>> a = np.empty(2)
    >>> a.fill(1)
    >>> a
    array([ 1.,  1.])

    """
    a[...] = value



"""
Random
~~~~~~

Random functions

"""
import bohrium as np
import numpy_force as numpy
from numpy_force.random import *
import operator
import functools
import datetime
import os
import sys
from bohrium import _bh
import math

from libc.stdint cimport uint64_t, uint32_t

cdef extern from "Random123/philox.h":
    struct r123array2x32:
        pass
    struct r123array1x32:
        pass
ctypedef r123array2x32 philox2x32_ctr_t
ctypedef r123array1x32 philox2x32_key_t
cdef extern from "Random123/philox.h":
    philox2x32_ctr_t philox2x32(philox2x32_ctr_t, philox2x32_key_t)
cdef union ctr_t:
    philox2x32_ctr_t c
    uint64_t ul
cdef union key_t:
    philox2x32_key_t k
    uint32_t ui

cdef extern from "Python.h":
    void*PyLong_AsVoidPtr(object)

cdef class RandomState:
    """
    RandomState(seed=None)

    Container for the Random123 pseudo-random number generator.

    `RandomState` exposes a number of methods for generating random numbers
    drawn from a variety of probability distributions. In addition to the
    distribution-specific arguments, each method takes a keyword argument
    `size` that defaults to ``None``. If `size` is ``None``, then a single
    value is generated and returned. If `size` is an integer, then a 1-D
    array filled with generated values is returned. If `size` is a tuple,
    then an array with that shape is filled and returned.

    Parameters
    ----------
    seed : int, optional
        Random seed initializing the pseudo-random number generator.
        Can be an integer or ``None`` (the default).
        If `seed` is ``None``, then `RandomState` will try to read data from
        ``/dev/urandom`` (or the Windows analogue) if available or seed from
        the clock otherwise.

    """
    cdef uint32_t key
    cdef uint64_t index
    cdef bint has_gauss
    cdef double gauss

    def __init__(self, seed=None):
        self.seed(seed)

    def seed(self, seed=None):
        """
        seed(seed=None)

        Seed the generator.

        This method is called when `RandomState` is initialized. It can be
        called again to re-seed the generator. For details, see `RandomState`.

        Parameters
        ----------
        seed : int or array_like, optional
            Seed for `RandomState`.

        See Also
        --------
        RandomState

        """
        if seed is None:
            try:
                self.key = numpy.uint32(hash(os.urandom(8)))
            except NotImplementedError:
                self.key = numpy.uint32(hash(datetime.datetime.now()))
        else:
            self.key = numpy.uint32(seed)
        self.index = 0;
        self.has_gauss = False

    def get_state(self):
        """
        get_state()

        Return a tuple representing the internal state of the generator.

        For more details, see `set_state`.

        Returns
        -------
        out : tuple(str, np.uint64, np.uint32)
            The returned tuple has the following items:

            1. the string 'Random123'.
            2. an integer ``index``.
            3. an integer ``key``.


        See Also
        --------
        set_state

        Notes
        -----
        `set_state` and `get_state` are not needed to work with any of the
        random distributions in Bohrium. If the internal state is manually altered,
        the user should know exactly what he/she is doing.

        """
        return ('Random123', self.index, self.key)

    def set_state(self, state):
        """
        set_state(state)

        Set the internal state of the generator from a tuple.

        For use if one has reason to manually (re-)set the internal state of the
        "Mersenne Twister"[1]_ pseudo-random number generating algorithm.

        Parameters
        ----------
        state : tuple(str, np.uint64, np.uint32)
            The returned tuple has the following items:

            1. the string 'Random123'.
            2. an integer ``index``.
            3. an integer ``key``.

        Returns
        -------
        out : None
            Returns 'None' on success.

        See Also
        --------
        get_state

        Notes
        -----
        `set_state` and `get_state` are not needed to work with any of the
        random distributions in Bohrium. If the internal state is manually altered,
        the user should know exactly what he/she is doing.
        """
        if len(state) != 3:
            raise ValueError("state must contain 3 elements")
        if state[0] != 'Random123':
            raise ValueError("algorithm must be 'Random123'")
        try:
            self.index = state[1]
            self.key = state[2]
        except TypeError:
            raise ValueError("state is not a valid Random123 state")

    def random123(self, size=None, bohrium=True):
        """
        New array of uniform pseudo numbers based on the random123 philox2x32 algorithm.
        NB: dtype is np.uint64 always

        Parameters
        ----------
        shape       : tuple of ints
        Defines the shape of the returned array of random floats.

        Returns
        -------
        out : Array of uniform pseudo numbers
        """
        cdef ctr_t ctr
        cdef ctr_t rnd
        cdef key_t key
        cdef uint64_t*array_data
        cdef long length
        cdef long i

        ctr.ul = self.index
        key.ui = self.key

        if size is None:
            length = 1
            rnd.c = philox2x32(ctr.c, key.k)
            ret = rnd.ul
        elif bohrium is False:
            ret = np.empty(size, dtype=np.uint64, bohrium=False)
            length = ret.size
            array_data = <uint64_t *> PyLong_AsVoidPtr(ret.ctypes.data)
            for i from 0 <= i < length:
                rnd.c = philox2x32(ctr.c, key.k)
                array_data[i] = rnd.ul
                ctr.ul += 1
        else:
            length = size if numpy.isscalar(size) else functools.reduce(operator.mul, size)
            ret = _bh.random123(length, self.index, self.key).reshape(size)
        self.index += length
        return ret

    def random_sample(self, size=None, dtype=float, bohrium=True):
        """
        Return random floats in the half-open interval [0.0, 1.0).

        Results are from the "continuous uniform" distribution over the
        stated interval.  To sample :math:`Unif[a, b), b > a` multiply
        the output of `random` by `(b-a)` and add `a`::

          (b - a) * random() + a

        Parameters
        ----------
        shape : int or tuple of ints, optional
            Defines the shape of the returned array of random floats. If None
            (the default), returns a single float.

        Returns
        -------
        out : float or ndarray of floats
            Array of random floats of shape `shape` (unless ``shape=None``, in which
            case a single float is returned).

        Examples
        --------
        >>> np.random.random()
        0.47108547995356098
        >>> type(np.random.random())
        <type 'float'>
        >>> np.random.random((5,))
        array([ 0.30220482,  0.86820401,  0.1654503 ,  0.11659149,  0.54323428])

        Three-by-two array of random numbers from [-5, 0):

        >>> 5 * np.random.random((3, 2)) - 5
        array([[-3.99149989, -0.52338984],
               [-2.99091858, -0.79479508],
               [-1.23204345, -1.75224494]])
        """
        dtype = np.dtype(dtype).type
        if not (dtype is np.float64 or dtype is np.float32):
            raise ValueError("dtype not supported for random_sample")
        #Generate random numbers as uint
        r_uint = self.random123(size, bohrium=bohrium)
        #Convert random numbers to float in the interval [0.0, 1.0).
        max_value = dtype(numpy.iinfo(numpy.uint64).max)
        if size is None:
            return numpy.dtype(dtype).type(r_uint) / max_value
        else:
            return np.array(r_uint, dtype=dtype, bohrium=bohrium) / max_value

    def tomaxint(self, size=None, bohrium=True):
        """
        tomaxint(size=None, bohrium=True)

        Random integers between 0 and ``sys.maxint``, inclusive.

        Return a sample of uniformly distributed random integers in the interval
        [0, ``sys.maxint``].

        Parameters
        ----------
        size : tuple of ints, int, optional
            Shape of output.  If this is, for example, (m,n,k), m*n*k samples
            are generated.  If no shape is specified, a single sample is
            returned.

        Returns
        -------
        out : ndarray
            Drawn samples, with shape `size`.

        See Also
        --------
        randint : Uniform sampling over a given half-open interval of integers.
        random_integers : Uniform sampling over a given closed interval of
            integers.

        Examples
        --------
        >>> RS = np.random.mtrand.RandomState() # need a RandomState object
        >>> RS.tomaxint((2,2,2))
        array([[[1170048599, 1600360186],
                [ 739731006, 1947757578]],
               [[1871712945,  752307660],
                [1601631370, 1479324245]]])
        >>> import sys
        >>> sys.maxint
        2147483647
        >>> RS.tomaxint((2,2,2)) < sys.maxint
        array([[[ True,  True],
                [ True,  True]],
               [[ True,  True],
                [ True,  True]]], dtype=bool)

        """
        r_uint = self.random123(size, bohrium) >> (64 - sys.maxint.bit_length())
        res = np.empty_like(r_uint, dtype=int)
        res[...] = r_uint
        return res

    def randint(self, low, high=None, size=None, dtype=int, bohrium=True):
        """
        randint(low, high=None, size=None, dtype=int, bohrium=True)

        Return random integers from `low` (inclusive) to `high` (exclusive).

        Return random integers from the "discrete uniform" distribution in the
        "half-open" interval [`low`, `high`). If `high` is None (the default),
        then results are from [0, `low`).

        Parameters
        ----------
        low : int
            Lowest (signed) integer to be drawn from the distribution (unless
            ``high=None``, in which case this parameter is the *highest* such
            integer).
        high : int, optional
            If provided, one above the largest (signed) integer to be drawn
            from the distribution (see above for behavior if ``high=None``).
        size : int or tuple of ints, optional
            Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
            ``m * n * k`` samples are drawn.  Default is None, in which case a
            single value is returned.

        Returns
        -------
        out : int or ndarray of ints
            `size`-shaped array of random integers from the appropriate
            distribution, or a single such random int if `size` not provided.

        See Also
        --------
        random.random_integers : similar to `randint`, only for the closed
            interval [`low`, `high`], and 1 is the lowest value if `high` is
            omitted. In particular, this other one is the one to use to generate
            uniformly distributed discrete non-integers.

        Examples
        --------
        >>> np.random.randint(2, size=10)
        array([1, 0, 0, 0, 1, 1, 0, 0, 1, 0])
        >>> np.random.randint(1, size=10)
        array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

        Generate a 2 x 4 array of ints between 0 and 4, inclusive:

        >>> np.random.randint(5, size=(2, 4))
        array([[4, 0, 2, 1],
               [3, 2, 2, 0]])

        """
        dtype = np.dtype(dtype).type
        if high is None:
            high = low
            low = 0
        if low >= high:
            raise ValueError("low >= high")
        diff = high - low
        if size is None:
            return dtype(dtype(self.random123(size, bohrium=bohrium) % diff) + low)
        else:
            return np.array(np.array(self.random123(size, bohrium=bohrium) % diff, dtype=dtype, bohrium=bohrium) + low,
                            dtype=dtype, bohrium=bohrium)

    def uniform(self, low=0.0, high=1.0, size=None, dtype=float, bohrium=True):
        """
        uniform(low=0.0, high=1.0, size=None, dtype=float, bohrium=True)

        Draw samples from a uniform distribution.

        Samples are uniformly distributed over the half-open interval
        ``[low, high)`` (includes low, but excludes high).  In other words,
        any value within the given interval is equally likely to be drawn
        by `uniform`.

        Parameters
        ----------
        low : float, optional
            Lower boundary of the output interval.  All values generated will be
            greater than or equal to low.  The default value is 0.
        high : float
            Upper boundary of the output interval.  All values generated will be
            less than high.  The default value is 1.0.
        size : int or tuple of ints, optional
            Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
            ``m * n * k`` samples are drawn.  Default is None, in which case a
            single value is returned.

        Returns
        -------
        out : ndarray
            Drawn samples, with shape `size`.

        See Also
        --------
        randint : Discrete uniform distribution, yielding integers.
        random_integers : Discrete uniform distribution over the closed
                          interval ``[low, high]``.
        random_sample : Floats uniformly distributed over ``[0, 1)``.
        random : Alias for `random_sample`.
        rand : Convenience function that accepts dimensions as input, e.g.,
               ``rand(2,2)`` would generate a 2-by-2 array of floats,
               uniformly distributed over ``[0, 1)``.

        Notes
        -----
        The probability density function of the uniform distribution is

        .. math:: p(x) = \\frac{1}{b - a}

        anywhere within the interval ``[a, b)``, and zero elsewhere.

        same as:
        random_sample(size) * (high - low) + low

        Examples
        --------
        Draw samples from the distribution:

        >>> s = np.random.uniform(-1,0,1000)

        All values are within the given interval:

        >>> np.all(s >= -1)
        True
        >>> np.all(s < 0)
        True

        Display the histogram of the samples, along with the
        probability density function:

        >>> import matplotlib.pyplot as plt
        >>> count, bins, ignored = plt.hist(s, 15, normed=True)
        >>> plt.plot(bins, np.ones_like(bins), linewidth=2, color='r')
        >>> plt.show()

        """
        dtype = np.dtype(dtype).type
        return self.random_sample(size=size, dtype=dtype, bohrium=bohrium) * dtype(high - low) + dtype(low)

    def rand(self, *args, dtype=float, bohrium=True):
        """
        rand(d0, d1, ..., dn, dtype=float, bohrium=True)

        Random values in a given shape.

        Create an array of the given shape and propagate it with
        random samples from a uniform distribution
        over ``[0, 1)``.

        Parameters
        ----------
        d0, d1, ..., dn : int, optional
            The dimensions of the returned array, should all be positive.
            If no argument is given a single Python float is returned.

        Returns
        -------
        out : ndarray, shape ``(d0, d1, ..., dn)``
            Random values.

        See Also
        --------
        random

        Notes
        -----
        This is a convenience function. If you want an interface that
        takes a shape-tuple as the first argument, refer to
        np.random.random_sample .

        Examples
        --------
        >>> np.random.rand(3,2)
        array([[ 0.14022471,  0.96360618],  #random
               [ 0.37601032,  0.25528411],  #random
               [ 0.49313049,  0.94909878]]) #random

        """
        if len(args) == 0:
            return self.random_sample(dtype=dtype, bohrium=bohrium)
        else:
            return self.random_sample(size=args, dtype=dtype, bohrium=bohrium)

    def randn(self, *args, dtype=float, bohrium=True):
        """
        randn(d0, d1, ..., dn, dtype=float, bohrium=True)

        Return a sample (or samples) from the "standard normal" distribution.

        If positive, int_like or int-convertible arguments are provided,
        `randn` generates an array of shape ``(d0, d1, ..., dn)``, filled
        with random floats sampled from a univariate "normal" (Gaussian)
        distribution of mean 0 and variance 1 (if any of the :math:`d_i` are
        floats, they are first converted to integers by truncation). A single
        float randomly sampled from the distribution is returned if no
        argument is provided.

        This is a convenience function.  If you want an interface that takes a
        tuple as the first argument, use `numpy.random.standard_normal` instead.

        Parameters
        ----------
        d0, d1, ..., dn : int, optional
            The dimensions of the returned array, should be all positive.
            If no argument is given a single Python float is returned.

        Returns
        -------
        Z : ndarray or float
            A ``(d0, d1, ..., dn)``-shaped array of floating-point samples from
            the standard normal distribution, or a single such float if
            no parameters were supplied.

        See Also
        --------
        random.standard_normal : Similar, but takes a tuple as its argument.

        Notes
        -----
        For random samples from :math:`N(\\mu, \\sigma^2)`, use:

        ``sigma * np.random.randn(...) + mu``

        Examples
        --------
        >>> np.random.randn()
        2.1923875335537315 #random

        Two-by-four array of samples from N(3, 6.25):

        >>> 2.5 * np.random.randn(2, 4) + 3
        array([[-4.49401501,  4.00950034, -1.81814867,  7.29718677],  #random
               [ 0.39924804,  4.68456316,  4.99394529,  4.84057254]]) #random

        """
        if len(args) == 0:
            return self.standard_normal(dtype=dtype, bohrium=bohrium)
        else:
            return self.standard_normal(size=args, dtype=dtype, bohrium=bohrium)

    def random_integers(self, low, high=None, size=None, dtype=int, bohrium=True):
        """
        random_integers(low, high=None, size=None, dtype=int, bohrium=True)

        Return random integers between `low` and `high`, inclusive.

        Return random integers from the "discrete uniform" distribution in the
        closed interval [`low`, `high`].  If `high` is None (the default),
        then results are from [1, `low`].

        Parameters
        ----------
        low : int
            Lowest (signed) integer to be drawn from the distribution (unless
            ``high=None``, in which case this parameter is the *highest* such
            integer).
        high : int, optional
            If provided, the largest (signed) integer to be drawn from the
            distribution (see above for behavior if ``high=None``).
        size : int or tuple of ints, optional
            Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
            ``m * n * k`` samples are drawn.  Default is None, in which case a
            single value is returned.

        Returns
        -------
        out : int or ndarray of ints
            `size`-shaped array of random integers from the appropriate
            distribution, or a single such random int if `size` not provided.

        See Also
        --------
        random.randint : Similar to `random_integers`, only for the half-open
            interval [`low`, `high`), and 0 is the lowest value if `high` is
            omitted.

        Notes
        -----
        To sample from N evenly spaced floating-point numbers between a and b,
        use::

          a + (b - a) * (np.random.random_integers(N) - 1) / (N - 1.)

        Examples
        --------
        >>> np.random.random_integers(5)
        4
        >>> type(np.random.random_integers(5))
        <type 'int'>
        >>> np.random.random_integers(5, size=(3.,2.))
        array([[5, 4],
               [3, 3],
               [4, 5]])

        Choose five random numbers from the set of five evenly-spaced
        numbers between 0 and 2.5, inclusive (*i.e.*, from the set
        :math:`{0, 5/8, 10/8, 15/8, 20/8}`):

        >>> 2.5 * (np.random.random_integers(5, size=(5,)) - 1) / 4.
        array([ 0.625,  1.25 ,  0.625,  0.625,  2.5  ])

        Roll two six sided dice 1000 times and sum the results:

        >>> d1 = np.random.random_integers(1, 6, 1000)
        >>> d2 = np.random.random_integers(1, 6, 1000)
        >>> dsums = d1 + d2

        Display results as a histogram:

        >>> import matplotlib.pyplot as plt
        >>> count, bins, ignored = plt.hist(dsums, 11, normed=True)
        >>> plt.show()

        """
        if high is None:
            high = low
            low = 1
        return self.randint(low, high + 1, size, dtype=dtype, bohrium=bohrium)

    # Complicated, continuous distributions:
    def standard_normal(self, size=None, dtype=float, bohrium=True):
        """
        standard_normal(size=None, dtype=float, bohrium=True)

        Returns samples from a Standard Normal distribution (mean=0, stdev=1).

        Parameters
        ----------
        size : int or tuple of ints, optional
            Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
            ``m * n * k`` samples are drawn.  Default is None, in which case a
            single value is returned.

        Returns
        -------
        out : float or ndarray
            Drawn samples.

        Examples
        --------
        >>> s = np.random.standard_normal(8000)
        >>> s
        array([ 0.6888893 ,  0.78096262, -0.89086505, ...,  0.49876311, #random
               -0.38672696, -0.4685006 ])                               #random
        >>> s.shape
        (8000,)
        >>> s = np.random.standard_normal(size=(3, 4, 2))
        >>> s.shape
        (3, 4, 2)

        """

        # Using basic Box-Muller transform
        dtype = np.dtype(dtype).type
        if not (dtype is np.float64 or dtype is np.float32):
            raise ValueError("dtype not supported for standart_normal")
        if size is None:
            if self.has_gauss:
                self.has_gauss = False
                return dtype(self.gauss)
            else:
                u1 = self.random_sample()
                u2 = self.random_sample()
                r = math.sqrt(-2. * math.log(u1))
                t = 2. * math.pi * u2
                z0 = r * math.cos(t)
                z1 = r * math.sin(t)
                self.gauss = z1
                self.has_gauss = True
                return dtype(z0)
        else:
            length = size if numpy.isscalar(size) else functools.reduce(operator.mul, size)
            hlength = length // 2 + length % 2
            u1 = self.random_sample(size=hlength, dtype=dtype, bohrium=bohrium)
            u2 = self.random_sample(size=hlength, dtype=dtype, bohrium=bohrium)
            r = np.sqrt(-2. * np.log(u1))
            t = 2. * math.pi * u2
            z0 = r * np.cos(t)
            z1 = r * np.sin(t)
            res = np.empty(hlength * 2, dtype=dtype, bohrium=bohrium)
            res[:hlength] = z0  # res[::2] = z0
            res[hlength:] = z1  # res[1::2] = z1
            return res[:length].reshape(size)

    def normal(self, loc=0.0, scale=1.0, size=None, dtype=float, bohrium=True):
        """
        normal(loc=0.0, scale=1.0, size=None, dtype=float, bohrium=True)

        Draw random samples from a normal (Gaussian) distribution.

        The probability density function of the normal distribution, first
        derived by De Moivre and 200 years later by both Gauss and Laplace
        independently [2]_, is often called the bell curve because of
        its characteristic shape (see the example below).

        The normal distributions occurs often in nature.  For example, it
        describes the commonly occurring distribution of samples influenced
        by a large number of tiny, random disturbances, each with its own
        unique distribution [2]_.

        Parameters
        ----------
        loc : float
            Mean ("centre") of the distribution.
        scale : float
            Standard deviation (spread or "width") of the distribution.
        size : int or tuple of ints, optional
            Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
            ``m * n * k`` samples are drawn.  Default is None, in which case a
            single value is returned.

        See Also
        --------
        scipy.stats.distributions.norm : probability density function,
            distribution or cumulative density function, etc.

        Notes
        -----
        The probability density for the Gaussian distribution is

        .. math:: p(x) = \\frac{1}{\\sqrt{ 2 \\pi \\sigma^2 }}
                         e^{ - \\frac{ (x - \\mu)^2 } {2 \\sigma^2} },

        where :math:`\\mu` is the mean and :math:`\\sigma` the standard deviation.
        The square of the standard deviation, :math:`\\sigma^2`, is called the
        variance.

        The function has its peak at the mean, and its "spread" increases with
        the standard deviation (the function reaches 0.607 times its maximum at
        :math:`x + \\sigma` and :math:`x - \\sigma` [2]_).  This implies that
        `numpy.random.normal` is more likely to return samples lying close to the
        mean, rather than those far away.

        References
        ----------
        .. [1] Wikipedia, "Normal distribution",
               http://en.wikipedia.org/wiki/Normal_distribution
        .. [2] P. R. Peebles Jr., "Central Limit Theorem" in "Probability, Random
               Variables and Random Signal Principles", 4th ed., 2001,
               pp. 51, 51, 125.

        Examples
        --------
        Draw samples from the distribution:

        >>> mu, sigma = 0, 0.1 # mean and standard deviation
        >>> s = np.random.normal(mu, sigma, 1000)

        Verify the mean and the variance:

        >>> abs(mu - np.mean(s)) < 0.01
        True

        >>> abs(sigma - np.std(s, ddof=1)) < 0.01
        True

        Display the histogram of the samples, along with
        the probability density function:

        >>> import matplotlib.pyplot as plt
        >>> count, bins, ignored = plt.hist(s, 30, normed=True)
        >>> plt.plot(bins, 1/(sigma * np.sqrt(2 * np.pi)) *
        ...                np.exp( - (bins - mu)**2 / (2 * sigma**2) ),
        ...          linewidth=2, color='r')
        >>> plt.show()

        """
        dtype = np.dtype(dtype).type
        scale = dtype(scale)
        loc = dtype(loc)
        if scale <= dtype(0):
            raise ValueError("scale <= 0")
        return self.standard_normal(size=size, dtype=dtype, bohrium=bohrium) * scale + loc

    def standard_exponential(self, size=None, dtype=float, bohrium=True):
        """
        standard_exponential(size=None, dtype=float, bohrium=True)

        Draw samples from the standard exponential distribution.

        `standard_exponential` is identical to the exponential distribution
        with a scale parameter of 1.

        Parameters
        ----------
        size : int or tuple of ints, optional
            Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
            ``m * n * k`` samples are drawn.  Default is None, in which case a
            single value is returned.

        Returns
        -------
        out : float or ndarray
            Drawn samples.

        Examples
        --------
        Output a 3x8000 array:

        >>> n = np.random.standard_exponential((3, 8000))

        """
        # We use -log(1-U) since U is [0, 1) */
        dtype = np.dtype(dtype).type
        return dtype(-1) * np.log(dtype(1) - self.random_sample(size=size, dtype=dtype, bohrium=bohrium))

    def exponential(self, scale=1.0, size=None, dtype=float, bohrium=True):
        """
        exponential(scale=1.0, size=None, dtype=float, bohrium=True)

        Exponential distribution.

        Its probability density function is

        .. math:: f(x; \\frac{1}{\\beta}) = \\frac{1}{\\beta} \\exp(-\\frac{x}{\\beta}),

        for ``x > 0`` and 0 elsewhere. :math:`\\beta` is the scale parameter,
        which is the inverse of the rate parameter :math:`\\lambda = 1/\\beta`.
        The rate parameter is an alternative, widely used parameterization
        of the exponential distribution [3]_.

        The exponential distribution is a continuous analogue of the
        geometric distribution.  It describes many common situations, such as
        the size of raindrops measured over many rainstorms [1]_, or the time
        between page requests to Wikipedia [2]_.

        Parameters
        ----------
        scale : float
            The scale parameter, :math:`\\beta = 1/\\lambda`.
        size : int or tuple of ints, optional
            Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
            ``m * n * k`` samples are drawn.  Default is None, in which case a
            single value is returned.

        References
        ----------
        .. [1] Peyton Z. Peebles Jr., "Probability, Random Variables and
               Random Signal Principles", 4th ed, 2001, p. 57.
        .. [2] "Poisson Process", Wikipedia,
               http://en.wikipedia.org/wiki/Poisson_process
        .. [3] "Exponential Distribution, Wikipedia,
               http://en.wikipedia.org/wiki/Exponential_distribution

        """
        dtype = np.dtype(dtype).type
        scale = dtype(scale)
        if scale <= dtype(0):
            raise ValueError("scale <= 0")
        return self.standard_exponential(size=size, dtype=dtype, bohrium=bohrium) * scale

    def random(self, shape, dtype, bohrium=True):
        """
        Return random numbers of 'dtype'.

        Parameters
        ----------
        shape : int or tuple of ints, optional
            Defines the shape of the returned array of random floats. If None
            (the default), returns a single float.

        Returns
        -------
        out : float or ndarray of floats
            Array of random floats of shape `shape` (unless ``shape=None``, in which
            case a single float is returned).
        """
        try:
            total = functools.reduce(operator.mul, shape)
        except TypeError:
            total = shape
            shape = (shape,)
        dtype = np.dtype(dtype).type
        if dtype is np.bool:
            res = self.random_integers(0, 1, shape, bohrium=bohrium)
        elif dtype in [np.int8, np.uint8]:
            res = self.random_integers(1, 3, shape, bohrium=bohrium)
        elif dtype is np.int16:
            res = self.random_integers(1, 5, shape, bohrium=bohrium)
        elif dtype is np.uint16:
            res = self.random_integers(1, 6, shape, bohrium=bohrium)
        elif dtype in [np.float32, np.float64]:
            res = self.random_sample(size=shape, bohrium=bohrium)
        elif dtype in [np.complex64, np.complex128]:
            res = self.random_sample(size=shape, bohrium=bohrium) + \
                  self.random_sample(size=shape, bohrium=bohrium) * 1j
        else:
            res = self.random_integers(1, 8, size=shape, bohrium=bohrium)
        if len(res.shape) == 0:  #Make sure scalars is arrays.
            res = np.asarray(res, bohrium=bohrium)
            res.shape = shape
        return np.asarray(res, dtype=dtype, bohrium=bohrium)

#The default random object
_inst = RandomState()
seed = _inst.seed
get_state = _inst.get_state
set_state = _inst.set_state
random_sample = _inst.random_sample
ranf = random = sample = random_sample
randint = _inst.randint
uniform = _inst.uniform
rand = _inst.rand
randn = _inst.randn
random_integers = _inst.random_integers
standard_normal = _inst.standard_normal
normal = _inst.normal
standard_exponential = _inst.standard_exponential
exponential = _inst.exponential



"""
Bohrium Contexts
================
"""
import sys
import os
from . import backend_messaging as messaging


class EnableBohrium:
    """Enable Bohrium within the context"""

    def __init__(self):
        # In order to avoid complications, we import common libraries BEFORE enabling Bohrium
        try:
            import matplotlib
            if os.environ.get("DISPLAY", "") == "":
                matplotlib.use('Agg')  # When no DISPLAY, we assume a headless matplotlib is used
            import matplotlib.pyplot
            import matplotlib.pylab
        except ImportError:
            pass
        try:
            import scipy
            import scipy.sparse
            import scipy.io
        except ImportError:
            pass
        try:
            import netCDF4
        except ImportError:
            pass
        try:
            import sklearn
            import sklearn.preprocessing
        except ImportError:
            pass

        # Let's save to real NumPy module
        self.__numpy = sys.modules['numpy']
        self.__numpy_random = sys.modules['numpy.random']
        self.__numpy_linalg = sys.modules['numpy.linalg']

        # Sub-module matlib has to be imported explicitly once in order to be available through bohrium
        try:
            import numpy.matlib
        except ImportError:
            pass

    def __enter__(self):
        import numpy
        import bohrium
        # Overwrite with Bohrium
        sys.modules['numpy_force'] = numpy
        sys.modules['numpy'] = bohrium
        sys.modules['numpy.random'] = bohrium.random
        sys.modules['numpy.linalg'] = bohrium.linalg

    def __exit__(self, *args):
        # Put NumPy back together
        sys.modules.pop('numpy_force', None)
        sys.modules['numpy'] = self.__numpy
        sys.modules['numpy.random'] = self.__numpy_random
        sys.modules['numpy.linalg'] = self.__numpy_linalg


class DisableBohrium:
    """Disable Bohrium within the context"""

    def __enter__(self):
        # Save current state
        import numpy
        self._numpy = sys.modules['numpy']
        self._numpy_random = sys.modules['numpy.random']
        self._numpy_linalg = sys.modules['numpy.linalg']
        # Make sure that numpy points to numpy (and not Bohrium)
        sys.modules['numpy'] = sys.modules.get("numpy_force", self._numpy)

    def __exit__(self, *args):
        # Load the state before entering context
        sys.modules['numpy'] = self._numpy
        sys.modules['numpy.random'] = self._numpy_random
        sys.modules['numpy.linalg'] = self._numpy_linalg


class Profiling:
    """Profiling the Bohrium backends within the context."""

    def __init__(self):
        pass

    def __enter__(self):
        messaging.statistic_enable_and_reset()

    def __exit__(self, *args):
        print(messaging.statistic())


class DisableGPU:
    """Disable the GPU backend within the context."""

    def __init__(self):
        pass

    def __enter__(self):
        messaging.gpu_disable()

    def __exit__(self, *args):
        messaging.gpu_enable()



#!/usr/bin/env python
"""
This module consist of bohrium.ndarray methods

The functions here serve as a means to determine whether a given
array is a numpy.ndarray or a bohrium.ndarray as well as moving
between the two "spaces".

--- License ---
This file is part of Bohrium and copyright (c) 2012 the Bohrium
http://bohrium.bitbucket.org

Bohrium is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

Bohrium is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the
GNU Lesser General Public License along with Bohrium.

If not, see <http://www.gnu.org/licenses/>.
"""
import sys
import numpy_force as numpy
cimport numpy as numpy

def check(ary):
    """Returns True if 'ary' is a Bohrium array"""

    try:
        #This will fail if the base is a NumPy array
        base = get_base(ary)
    except AttributeError:
        base = ary
    from . import _bh  #We import locally in order to avoid import cycle
    return type(ary) is _bh.ndarray

def check_biclass_np_over_bh(ary):
    """Returns True if 'ary' is a NumPy view with a Bohrium base array"""

    if not isinstance(ary, numpy.ndarray):
        return False
    try:
        if not check(get_base(ary)):
            return False
    except AttributeError:
        return False

    return not check(ary)

def check_biclass_bh_over_np(ary):
    """Returns True if 'ary' is a Bohrium view with a NumPy base array"""

    return check(ary) and not check(get_base(ary))

def fix_biclass(ary):
    """
    Makes sure that when 'ary' or its base is a Bohrium array, both of them are.
    """

    if not isinstance(ary, numpy.ndarray):
        return ary

    if check_biclass_np_over_bh(ary):
        return ary.view(type(get_base(ary)))
    elif check_biclass_bh_over_np(ary):
        from . import array_create
        ary = array_create.array(ary, bohrium=False)
        return array_create.array(ary, bohrium=True)
    else:
        return ary

def fix_biclass_wrapper(func):
    """
    Function decorator that makes sure that the function doesn't reads or writes biclass arrays
    """

    if hasattr(func, "_wrapped_fix_biclass"):
        return func

    def inner(*args, **kwargs):
        """Invokes 'func' and strips "biclass" from the result."""

        # Normally, we checks all arguments for biclass arrays, but this can be disabled individually
        # by setting the function keyword argument 'fix_biclass' to False
        if kwargs.get("fix_biclass", True) and sys.version_info[0] < 3:
            args = [fix_biclass(a) for a in args]
        if "fix_biclass" in kwargs:
            del kwargs["fix_biclass"]
        ret = func(*args, **kwargs)
        return fix_biclass(ret)

    try:
        #Flag that this function has been handled
        setattr(inner, "_wrapped_fix_biclass", True)
    except:  #In older versions of cython, this is not possible
        pass
    return inner

def get_base(ary):
    """Get the final base array of 'ary'."""

    if ary.base is None:
        return ary
    else:
        base = ary.base
        while base.base is not None:
            base = base.base
        return base

def is_base(ary):
    """Return True when 'ary' is a base array."""

    base = get_base(ary)
    return base is ary



"""
Masking routines
===========================

"""
import warnings
from . import array_create
from . import bhary
from . import reorganization
from . import array_manipulation
from . import ufuncs
from . import summations
import numpy_force as numpy
from .bhary import fix_biclass_wrapper


@fix_biclass_wrapper
def where(condition, x=None, y=None):
    """
    where(condition, [x, y])

    Return elements, either from `x` or `y`, depending on `condition`.

    If only `condition` is given, return ``condition.nonzero()``.

    Parameters
    ----------
    condition : array_like, bool
        When True, yield `x`, otherwise yield `y`.
    x, y : array_like, optional
        Values from which to choose. `x` and `y` need to have the same
        shape as `condition`.

    Returns
    -------
    out : ndarray or tuple of ndarrays
        If both `x` and `y` are specified, the output array contains
        elements of `x` where `condition` is True, and elements from
        `y` elsewhere.

        If only `condition` is given, return the tuple
        ``condition.nonzero()``, the indices where `condition` is True.

    See Also
    --------
    nonzero, choose

    Notes
    -----
    If `x` and `y` are given and input arrays are 1-D, `where` is
    equivalent to::

        [xv if c else yv for (c,xv,yv) in zip(condition,x,y)]

    Examples
    --------
    >>> np.where([[True, False], [True, True]],
    ...          [[1, 2], [3, 4]],
    ...          [[9, 8], [7, 6]])
    array([[1, 8],
           [3, 4]])

    >>> np.where([[0, 1], [1, 0]])
    (array([0, 1]), array([1, 0]))

    >>> x = np.arange(9.).reshape(3, 3)
    >>> np.where( x > 5 )
    (array([2, 2, 2]), array([0, 1, 2]))
    >>> x[np.where( x > 3.0 )]               # Note: result is 1D.
    array([ 4.,  5.,  6.,  7.,  8.])
    >>> np.where(x < 5, x, -1)               # Note: broadcasting.
    array([[ 0.,  1.,  2.],
           [ 3.,  4., -1.],
           [-1., -1., -1.]])

    Find the indices of elements of `x` that are in `goodvalues`.

    >>> goodvalues = [3, 4, 7]
    >>> ix = np.in1d(x.ravel(), goodvalues).reshape(x.shape)
    >>> ix
    array([[False, False, False],
           [ True,  True, False],
           [False,  True, False]], dtype=bool)
    >>> np.where(ix)
    (array([1, 1, 2]), array([0, 1, 1]))

    """
    if x is None or y is None:
        warnings.warn("Bohrium only supports where() when 'x' and 'y' are specified", stacklevel=2)
        return numpy.where(condition)

    if not (bhary.check(condition) or bhary.check(x) or bhary.check(y)):
        return numpy.where(condition, x, y)

    # Make sure that non-scalars are Bohrium arrays
    if numpy.isscalar(condition):
        condition = bool(condition)
    else:
        condition = array_create.array(condition).astype("bool")

    if not numpy.isscalar(x):
        x = array_create.array(x)

    if not numpy.isscalar(y):
        y = array_create.array(y)

    # Shortcut if all arguments are scalars
    if all(numpy.isscalar(k) or k.size == 1 for k in (x, y, condition)):
        return x if condition else y

    # Find appropriate output type
    array_types = []
    scalar_types = []
    for v in (x, y):
        if numpy.isscalar(v):
            scalar_types.append(type(v))
        else:
            array_types.append(v.dtype)
    out_type = numpy.find_common_type(array_types, scalar_types)

    # Shortcut if input arrays are finite
    if ufuncs.isfinite(x).all() and ufuncs.isfinite(y).all():
        if numpy.isscalar(condition):
            res = condition * x + (not condition) * y
        else:
            res = condition * x + ufuncs.logical_not(condition) * y
        if numpy.isscalar(res):
            return out_type(res)
        else:
            return res.astype(out_type)

    # General case: use fancy indexing
    (condition, x, y), newshape = array_manipulation.broadcast_arrays(condition, x, y)
    ret = array_create.zeros(newshape, dtype=out_type)
    ret[condition] = x if numpy.isscalar(x) else x[condition]
    ret[~condition] = y if numpy.isscalar(y) else y[~condition]
    return ret


def masked_get(ary, bool_mask):
    """
    Get the elements of 'ary' specified by 'bool_mask'.
    """

    return ary[reorganization.nonzero(bool_mask)]


def masked_set(ary, bool_mask, value):
    """
    Set the 'value' into 'ary' at the location specified through 'bool_mask'.
    """

    if numpy.isscalar(value) and ufuncs.isfinite(value):
        ary *= ~bool_mask
        ary += bool_mask * value
    else:
        ary[reorganization.nonzero(bool_mask)] = value



"""
Interop NumPy
~~~~~~~~~~~~~
"""
from .bhary import get_base
from bohrium_api import stack_info


def get_array(bh_ary):
    """Return a NumPy array wrapping the memory of the Bohrium array `ary`.
    
    Parameters
    ----------
    bh_ary : ndarray (Bohrium array)
        Must be a Bohrium base array
    
    Returns
    -------
    out : ndarray (regular NumPy array)

    Notes
    -----
    Changing or deallocating `bh_ary` invalidates the returned NumPy array!

    """

    if stack_info.is_proxy_in_stack():
        raise RuntimeError("Cannot directly access array data through a proxy.")

    if get_base(bh_ary) is not bh_ary:
        raise RuntimeError('`bh_ary` must be a base array and not a view')
    assert(bh_ary.bhc_mmap_allocated)

    return bh_ary._numpy_wrapper()



from .. import array_create
from .._util import dtype_equal
from ..interop_numpy import get_array
from ..bhary import get_base
from .._bh import flush

import cython
import numpy_force as np

from libc.stdlib cimport abort, malloc, free
cimport numpy as cnp
ctypedef cnp.uint64_t uint64
ctypedef cnp.int64_t int64

@cython.boundscheck(False) # turn off bounds-checking
@cython.cdivision(True) # turn off division-by-zero checking
cdef _count(uint64[:] x, uint64[:] out):
    cdef int num_threads, thds_id
    cdef uint64 i, start, end
    cdef uint64* local_histo

    with nogil:
        for i in range(x.shape[0]):
            out[x[i]] += 1


@cython.boundscheck(False) # turn off bounds-checking
@cython.cdivision(True) # turn off division-by-zero checking
cdef _count_int_weights(uint64[:] x, int64[:] w, int64[:] out):
    cdef int num_threads, thds_id
    cdef uint64 i, start, end
    cdef int64* local_histo

    with nogil:
        for i in range(x.shape[0]):
            out[x[i]] += w[i]


@cython.boundscheck(False) # turn off bounds-checking
@cython.cdivision(True) # turn off division-by-zero checking
cdef _count_float_weights(uint64[:] x, double[:] w, double[:] out):
    cdef int num_threads, thds_id
    cdef uint64 i, start, end
    cdef double* local_histo

    with nogil:
        for i in range(x.shape[0]):
            out[x[i]] += w[i]


def bincount_cython(x, weights=None, minlength=None):
    """Cython implementation of `bincount()`"""

    x_max = int(x.max())
    if x_max < 0:
        raise RuntimeError("bincount(): first argument must be a 1 dimensional, non-negative int array")
    if minlength is not None:
        x_max = max(x_max, minlength)

    x = array_create.array(x, dtype=np.uint64)
    flush()
    if weights is None:
        ret = array_create.zeros((x_max+1, ), dtype=x.dtype)
        _count(get_array(x), get_array(ret))
    else:
        if np.issubdtype(weights.dtype, np.integer):
            weights = array_create.array(weights, dtype=np.int64)
            ret = array_create.zeros((x_max+1, ), dtype=weights.dtype)
            _count_int_weights(get_array(x), get_array(weights), get_array(ret))
        elif dtype_equal(weights.dtype, np.float32) or dtype_equal(weights.dtype, np.float64):
            weights = array_create.array(weights, dtype=np.float64)
            ret = array_create.zeros((x_max+1, ), dtype=weights.dtype)
            _count_float_weights(get_array(x), get_array(get_base(weights)), get_array(ret))
        else:
            raise RuntimeError("bincount(): weights has unsupported dtype (%s)" % weights.dtype)
    return ret



#!/usr/bin/env python
"""
=========================
NumPy ufunc encapsulation
=========================
"""

from __future__ import print_function
import sys
import os
import warnings
from . import _util
from . import array_create
from . import loop
import numpy_force as np
from bohrium_api import _info, stack_info
from .numpy_backport import as_strided
from ._util import dtype_equal
from .bhary import fix_biclass_wrapper
from . import bhary
from .array_manipulation import broadcast_arrays


@fix_biclass_wrapper
def extmethod(name, out, in1, in2):
    from . import _bh
    _bh.extmethod(name, (out, in1, in2))

def setitem(ary, loc, value):
    """
    Set the 'value' into 'ary' at the location specified through 'loc'.
    'loc' can be a scalar or a slice object, or a tuple thereof
    """

    if not isinstance(loc, tuple):
        loc = (loc,)

    # Let's try to convert non-arrays and non-scalars to an array
    # e.g. converting a python list to an array
    if not (bhary.check(value) or np.isscalar(value)):
        value = array_create.array(value)

    # Lets make sure that not all dimensions are indexed by integers
    loc = list(loc)
    if len(loc) == ary.ndim and all((np.isscalar(s) for s in loc)):
        # 'slice' doesn't support negative start index
        if loc[0] < 0:
            loc[0] += ary.shape[0]
        loc[0] = slice(loc[0], loc[0] + 1)

    # Copy the 'value' to 'ary' using the 'loc'
    if ary.ndim == 0:
        assign(value, ary)
    else:
        assign(value, ary[tuple(loc)])

def overlap_conflict(out, *inputs):
    """
    Return True when there is a possible memory conflict between the output
    and the inputs.


    :param Mixed out: Array in the role of being written to.
    :param Mixed inputs: Arrays in the role being read from.
    :returns: True in case of conflict.
    :rtype: bool
    """
    from . import _bh

    for i in inputs:
        if not np.isscalar(i):
            if np.may_share_memory(out, i) and not _bh.same_view(out, i):
                return True
    return False

@fix_biclass_wrapper
def assign(ary, out):
    """Copy data from array 'ary' to 'out'"""

    from . import _bh

    if not np.isscalar(ary):
        (ary, out) = broadcast_arrays(ary, out)[0]
        # We ignore self assignments
        if _bh.same_view(ary, out):
            return

    # Assigning empty arrays doesn't do anything
    if hasattr(ary, "size"):
        if ary.size == 0:
            return
    if hasattr(out, "size"):
        if out.size == 0:
            return

    # We use a tmp array if the in-/out-put has memory conflicts
    if overlap_conflict(out, ary):
        tmp = array_create.empty_like(out)
        assign(ary, tmp)
        return assign(tmp, out)

    if bhary.check(out):
        _bh.ufunc(UFUNCS["identity"].info['id'], (out, ary))
    else:
        if bhary.check(ary):
            if "BH_SYNC_WARN" in os.environ:
                import warnings
                warnings.warn("BH_SYNC_WARN: Copying the array to NumPy", RuntimeWarning, stacklevel=2)
            ary = ary.copy2numpy()
        out[...] = ary


class Ufunc(object):
    def __init__(self, info):
        """A Bohrium Universal Function"""
        self.info = info
        if sys.version_info.major >= 3:
            self.__name__ = info['name']
        else:
            # Scipy complains if '__name__' is unicode
            self.__name__ = info['name'].encode('latin_1')

    def __str__(self):
        return "<bohrium Ufunc '%s'>" % self.info['name']

    def __call__(self, *args, **kwargs):
        from . import _bh
        args = list(args)

        # Check number of array arguments
        if len(args) != self.info['nop'] and len(args) != self.info['nop'] - 1:
            raise ValueError("invalid number of array arguments")

        # Let's make sure that 'out' is always a positional argument
        try:
            out = kwargs['out']
            del kwargs['out']
            if len(args) == self.info['nop']:
                raise ValueError("cannot specify 'out' as both a positional and keyword argument")
            args.append(out)
        except KeyError:
            pass

        # We ignore the "casting" argument. In Bohrium, we always use unsafe casting
        try:
            out = kwargs['casting']
            del kwargs['casting']
        except KeyError:
            pass

        # We do not support NumPy's exotic arguments
        for k, val in kwargs.items():
            if val is not None:
                raise ValueError("Bohrium ufuncs doesn't support the '%s' argument" % str(k))

        # We let NumPy handle bharray views of ndarray base arrays
        for i in range(len(args)):
            if bhary.check_biclass_bh_over_np(args[i]):
                args[i] = as_strided(bhary.get_base(args[i]), shape=args[i].shape, strides=args[i].strides)
                np_ufunc = eval("np.%s" % self.info['name'])
                return np_ufunc(*args)

        # Makes sure that `args` are either bohrium arrays or scalars
        for i in range(len(args)):
            if not np.isscalar(args[i]) and not bhary.check(args[i]):
                args[i] = array_create.array(args[i])

        # Broadcast the args
        (bargs, out_shape) = broadcast_arrays(*args)

        # Pop the output from the 'bargs' list
        out = None
        if len(args) == self.info['nop']:
            out = args.pop()
            if out_shape != out.shape:
                raise ValueError("non-broadcastable output operand with shape %s "
                                 "doesn't match the broadcast shape %s" %
                                 (str(args[-1].shape), str(out.shape)))

        # We use a tmp array if the in-/out-put has memory conflicts
        if out is not None:
            if overlap_conflict(out, *args):
                tmp = self.__call__(*args, **kwargs)
                assign(tmp, out)
                return out

        # Copy broadcasted array back to 'args' excluding scalars
        for i in range(len(args)):
            if not np.isscalar(args[i]):
                args[i] = bargs[i]

        if any([bhary.check(a) for a in args]):
            if out is not None and not bhary.check(out):
                raise NotImplementedError("For now, the output must be a Bohrium array when the input arrays are")
        elif not bhary.check(out):
            # All operands are regular NumPy arrays
            func = eval("np.%s" % self.info['name'])
            if out is not None:
                args.append(out)
            return func(*args)

        if len(args) > 2:
            raise ValueError("Bohrium do not support ufunc with more than two inputs")

        # Convert 0-dim ndarray into scalars
        for i in range(len(args)):
            if not np.isscalar(args[i]):
                base = bhary.get_base(args[i])
                if not bhary.check(base):
                    if np.isscalar(base):
                        args[i] = base
                    elif base.ndim == 0:
                        args[i] = base.item()

        # Find the type signature
        (out_dtype, in_dtype) = _util.type_sig(self.info['name'], args)

        # Convert dtype of all inputs to match the function type signature
        for i in range(len(args)):
            if not dtype_equal(args[i], in_dtype):
                if np.isscalar(args[i]):
                    args[i] = in_dtype.type(args[i])
                else:
                    tmp = array_create.empty_like(args[i], dtype=in_dtype)
                    tmp[...] = args[i]
                    args[i] = tmp

        # Insert the output array
        if out is None or not dtype_equal(out_dtype, out.dtype):
            # Create a new empty array
            dynamic_out = array_create.empty(out_shape, out_dtype)

            # If the views within the ufunc has dynamic changes, then the output array must
            # also be dynamic. Used in regard to iterators within do_while loops.
            out_dvi = loop.DynamicViewInfo({}, dynamic_out.shape, dynamic_out.strides)
            for a in args:
                if bhary.check(a) and a.bhc_dynamic_view_info:
                    a_dvi = a.bhc_dynamic_view_info
                    for dim in a_dvi.dims_with_changes():
                        if a_dvi.dim_shape_change(dim) == 0:
                            continue
                        if out_dvi.has_changes_in_dim(dim):
                            # If the argument views have different shape change  in the same dimension,
                            # the dynamic shape of the temporary view can not be guessed and therefore
                            # results in an error
                            assert(out_dvi.dim_shape_change(dim) == a_dvi.dim_shape_change(dim))
                            a_reset = dim in a_dvi.resets
                            out_reset = dim in out_dvi.resets
                            assert((a_reset and out_reset) or ((not a_reset) and (not out_reset)))
                            if a_reset and out_reset:
                                assert(out_dvi.resets[dim] == a_dvi.resets[dim])
                        else:
                            for (_,shape_change, step_delay, _, _) in a_dvi.changes_in_dim(dim):
                                # No reason to add a change of 0 in the dimension
                                if shape_change == 0:
                                    continue
                                else:
                                    out_dvi.add_dynamic_change(dim, 0,shape_change, step_delay)

            # Only add dynamic view info if it actually contains changes
            if out_dvi.has_changes():
                dynamic_out.bhc_dynamic_view_info = out_dvi

            args.insert(0, dynamic_out)
        else:
            args.insert(0, out)

        # Call the bhc API
        if self.info['name'] == "power" and np.isscalar(args[2]) and args[2] == 2:  # A simple "power" optimization
            # Replace power of 2 with a multiplication
            _bh.ufunc(UFUNCS["multiply"].info['id'], (args[0], args[1], args[1]))
        else:
            _bh.ufunc(self.info['id'], args)

        if out is None or dtype_equal(out_dtype, out.dtype):
            return args[0]
        else:
            # We need to convert the output type before returning
            assign(args[0], out)
            return out

    @fix_biclass_wrapper
    def reduce(self, ary, axis=0, out=None):
        """
        A Bohrium Reduction
        Reduces `ary`'s dimension by len('axis'), by applying ufunc along the
        axes in 'axis'.

        Let :math:`ary.shape = (N_0, ..., N_i, ..., N_{M-1})`.  Then
        :math:`ufunc.reduce(ary, axis=i)[k_0, ..,k_{i-1}, k_{i+1}, .., k_{M-1}]` =
        the result of iterating `j` over :math:`range(N_i)`, cumulatively applying
        ufunc to each :math:`ary[k_0, ..,k_{i-1}, j, k_{i+1}, .., k_{M-1}]`.
        For a one-dimensional array, reduce produces results equivalent to:

          r = op.identity # op = ufunc
          for i in range(len(A)):
              r = op(r, A[i])
          return r

        For example, add.reduce() is equivalent to sum().

        Parameters
        ----------
        ary : array_like
            The array to act on.
        axis : None or int or tuple of ints, optional
            Axis or axes along which a reduction is performed.
            The default (`axis` = 0) is perform a reduction over the first
            dimension of the input array. `axis` may be negative, in
            which case it counts from the last to the first axis.

            .. versionadded:: 1.7.0

            If this is `None`, a reduction is performed over all the axes.
            If this is a tuple of ints, a reduction is performed on multiple
            axes, instead of a single axis or all the axes as before.

            For operations which are either not commutative or not associative,
            doing a reduction over multiple axes is not well-defined. The
            ufuncs do not currently raise an exception in this case, but will
            likely do so in the future.
        out : ndarray, optional
            A location into which the result is stored. If not provided, a
            freshly-allocated array is returned.

        Returns
        -------
        r : ndarraout      The reduced array. If `out` was supplied, `r` is a reference to it.

        Examples
        --------
        >>> np.multiply.reduce([2,3,5])
        30

        A multi-dimensional array example:

        >>> X = np.arange(8).reshape((2,2,2))
        >>> X
        array([[[0, 1],
                [2, 3]],
               [[4, 5],
                [6, 7]]])
        >>> np.add.reduce(X, 0)
        array([[ 4,  6],
               [ 8, 10]])
        >>> np.add.reduce(X) # confirm: default axis value is 0
        array([[ 4,  6],
               [ 8, 10]])
        >>> np.add.reduce(X, 1)
        array([[ 2,  4],
               [10, 12]])
        >>> np.add.reduce(X, 2)
        array([[ 1,  5],
               [ 9, 13]])
        """
        from . import _bh

        if out is not None:
            if bhary.check(out):
                if not bhary.check(ary):
                    ary = array_create.array(ary)
            else:
                if bhary.check(ary):
                    ary = ary.copy2numpy()

        # Let NumPy handle NumPy array reductions
        if not bhary.check(ary):
            func = eval("np.%s.reduce" % self.info['name'])
            return func(ary, axis=axis, out=out)

        # Make sure that 'axis' is a list of dimensions to reduce
        if axis is None:
            # We reduce all dimensions
            axis = range(ary.ndim)
        elif np.isscalar(axis):
            # We reduce one dimension
            axis = [axis]
        else:
            # We reduce multiple dimensions
            axis = list(axis)

        # Check for out of bounds and convert negative axis values
        if len(axis) > ary.ndim:
            raise ValueError("number of 'axes' to reduce is out of bounds")
        for i in range(len(axis)):
            if axis[i] < 0:
                axis[i] = ary.ndim + axis[i]
            if axis[i] >= ary.ndim:
                raise ValueError("'axis' is out of bounds")

        # No axis should be reduced multiple times
        if len(axis) != len(set(axis)):
            raise ValueError("duplicate value in 'axis'")

        # When reducing booleans numerically, we count the number of True values
        if (not self.info['name'].startswith("logical")) and dtype_equal(ary, np.bool):
            ary = array_create.array(ary, dtype=np.uint64)

        if len(axis) == 1:
            # One axis reduction we can handle directly
            axis = axis[0]

            # Find the output shape
            if ary.ndim == 1:
                shape = []
            else:
                shape = tuple(s for i, s in enumerate(ary.shape) if i != axis)
                if out is not None and out.shape != shape:
                    raise ValueError("output dimension mismatch expect " \
                                     "shape '%s' got '%s'" % (shape, out.shape))

            tmp = array_create.empty(shape, dtype=ary.dtype)

            # NumPy compatibility: when the axis dimension size is zero NumPy just returns the neutral value
            if ary.shape[axis] == 0:
                tmp[...] = getattr(getattr(np, self.info['name']), "identity")
            else:
                _bh.ufunc(_info.op["%s_reduce" % self.info['name']]['id'], (tmp, ary, np.int64(axis)))

            if out is not None:
                out[...] = tmp
            else:
                out = tmp

            return out
        else:
            # Let's sort the axis indexes by their stride
            # We use column major when a GPU is in the stack
            column_major = stack_info.is_opencl_in_stack() or stack_info.is_cuda_in_stack()
            strides = []
            for i, s in enumerate(ary.strides):
                if i in axis:
                    strides.append((i, abs(s)))
            axis = [i[0] for i in sorted(strides, key=lambda x: x[1], reverse=column_major)]

            # Let's reduce the first axis
            ary = self.reduce(ary, axis[0])
            # Then we reduce the rest of the axes and remember to correct the axis values
            axis = [i if i < axis[0] else i-1 for i in axis[1:]]
            ary = self.reduce(ary, axis)
            # Finally, we may have to copy the result to 'out'
            if out is not None:
                out[...] = ary
            else:
                out = ary
            return out

    @fix_biclass_wrapper
    def accumulate(self, ary, axis=0, out=None):
        """
        accumulate(ary, axis=0, out=None)

        Accumulate the result of applying the operator to all elements.

        For a one-dimensional array, accumulate produces results equivalent to::

          r = np.empty(len(A))
          t = op.identity        # op = the ufunc being applied to A's  elements
          for i in range(len(A)):
              t = op(t, A[i])
              r[i] = t
          return r

        For example, add.accumulate() is equivalent to np.cumsum().

        For a multi-dimensional array, accumulate is applied along only one
        axis (axis zero by default; see Examples below) so repeated use is
        necessary if one wants to accumulate over multiple axes.

        Parameters
        ----------
        ary : array_like
            The array to act on.
        axis : int, optional
            The axis along which to apply the accumulation; default is zero.
        out : ndarray, optional
            A location into which the result is stored. If not provided a
            freshly-allocated array is returned.

        Returns
        -------
        r : ndarray
            The accumulated values. If `out` was supplied, `r` is a reference to
            `out`.

        Examples
        --------
        1-D array examples:

        >>> np.add.accumulate([2, 3, 5])
        array([ 2,  5, 10])
        >>> np.multiply.accumulate([2, 3, 5])
        array([ 2,  6, 30])

        2-D array examples:

        >>> I = np.eye(2)
        >>> I
        array([[ 1.,  0.],
               [ 0.,  1.]])

        Accumulate along axis 0 (rows), down columns:

        >>> np.add.accumulate(I, 0)
        array([[ 1.,  0.],
               [ 1.,  1.]])
        >>> np.add.accumulate(I) # no axis specified = axis zero
        array([[ 1.,  0.],
               [ 1.,  1.]])

        Accumulate along axis 1 (columns), through rows:

        >>> np.add.accumulate(I, 1)
        array([[ 1.,  1.],
               [ 0.,  1.]])
        """
        from . import _bh

        if out is not None:
            if bhary.check(out):
                if not bhary.check(ary):
                    ary = array_create.array(ary)
            else:
                if bhary.check(ary):
                    ary = ary.copy2numpy()
            if out.shape != ary.shape:
                raise ValueError("output dimension mismatch expect " \
                                 "shape '%s' got '%s'" % (ary.shape, out.shape))

        # Check for out of bounds and convert negative axis values
        if axis < 0:
            axis = ary.ndim + axis
        if axis >= ary.ndim:
            raise ValueError("'axis' is out of bounds")

        # Let NumPy handle NumPy array accumulate
        if not bhary.check(ary):
            func = eval("np.%s.accumulate" % self.info['name'])
            return func(ary, axis=axis, out=out)

        # When reducing booleans numerically, we count the number of True values
        if dtype_equal(ary, np.bool):
            ary = array_create.array(ary, dtype=np.uint64)

        if out is None:
            out = array_create.empty(ary.shape, dtype=ary.dtype)

        _bh.ufunc(_info.op["%s_accumulate" % self.info['name']]['id'], (out, ary, np.int64(axis)))
        return out


#
# Expose all ufuncs at the module-level.
#
# After the following is executed, all ufuncs will be available as
# object instances of the Ufunc class via the dict of all ufuncs (UFUNCS)
# and via their individual names such as "negative", "identity", "add" etc.

# Expose via UFUNCS
UFUNCS = {}
for op in _info.op.values():
    if op['elementwise']:
        f = Ufunc(op)
        UFUNCS[f.info['name']] = f

# 'bh_divide' refers to how Bohrium divide, which is like division in C/C++
# We needs this reference because Python v3 uses "true" division
UFUNCS["bh_divide"] = UFUNCS["divide"]


# NOTE: We have to add ufuncs that doesn't map to Bohrium operations directly
#       such as "negative" which can be done like below.
class Negative(Ufunc):
    @fix_biclass_wrapper
    def __call__(self, ary, out=None):
        if out is None:
            return -1 * ary
        else:
            out[...] = -1 * ary
            return out


UFUNCS["negative"] = Negative({'name': 'negative'})


class TrueDivide(Ufunc):
    @fix_biclass_wrapper
    def __call__(self, a1, a2, out=None):
        if _util.dtype_is_float(a1) or _util.dtype_is_float(a2):
            # Floating points automatically use Bohrium division
            ret = UFUNCS["bh_divide"](a1, a2)
        else:
            if a1.dtype.itemsize > 4 or a2.dtype.itemsize > 4:
                dtype = np.float64
            else:
                dtype = np.float32
            ret = array_create.array(a1, dtype=dtype) / array_create.array(a2, dtype=dtype)
        if out is None:
            return ret
        else:
            out[...] = ret
            return out


UFUNCS["true_divide"] = TrueDivide({'name': 'true_divide'})


class FloorDivide(Ufunc):
    @fix_biclass_wrapper
    def __call__(self, a1, a2, out=None):
        if _util.dtype_is_float(a1) or _util.dtype_is_float(a2):
            ret = UFUNCS["floor"](a1 / a2)
        else:
            # Integers automatically use Bohrium division
            ret = UFUNCS["bh_divide"](a1, a2)
        if out is None:
            return ret
        else:
            out[...] = ret
            return out


UFUNCS["floor_divide"] = FloorDivide({'name': 'floor_divide'})

# Python v3 uses "true" division
if sys.version_info.major >= 3:
    UFUNCS["divide"] = UFUNCS["true_divide"]

# Expose via their name.
for name, ufunc in UFUNCS.items():
    exec ("%s = ufunc" % name)

# We do not want to expose a function named "ufunc"
del ufunc

def _handle__array_ufunc__(self, ufunc, method, *inputs, **kwargs):
    """
    - *self* is the array the method `__array_ufunc__` was called from.
    - *ufunc* is the ufunc object that was called.
    - *method* is a string indicating how the Ufunc was called, either
      ``"__call__"`` to indicate it was called directly, or one of its
      :ref:`methods<ufuncs.methods>`: ``"reduce"``, ``"accumulate"``,
      ``"reduceat"``, ``"outer"``, or ``"at"``.
    - *inputs* is a tuple of the input arguments to the ``ufunc``
    - *kwargs* contains any optional or keyword arguments passed to the
      function. This includes any ``out`` arguments, which are always
      contained in a tuple.
    """

    if method == '__call__' and ufunc.__name__ in UFUNCS:
        # if `out` is set, it must be a single Bohrium array
        if 'out' not in kwargs or len(kwargs['out']) == 1 and bhary.check(kwargs['out'][0]):
            return UFUNCS[ufunc.__name__](*inputs, **kwargs)
        else:
            warnings.warn("Bohrium does not support regular numpy arrays as output, it will be handled by "
                          "the original NumPy.", UserWarning, 1)
    else:
        warnings.warn("Bohrium does not support ufunc `%s` it will be handled by "
                      "the original NumPy." % ufunc.__name__, UserWarning, 1)
    np_inputs = []
    for i in inputs:
        if bhary.check(i):
            np_inputs.append(i.copy2numpy())
        else:
            np_inputs.append(i)
    return getattr(np, ufunc.__name__)(*np_inputs, **kwargs)



"""
Summations and products
~~~~~~

Common linear algebra functions

"""
import warnings
from . import ufuncs
from . import array_create
from . import bhary
from . import reorganization
from . import array_manipulation
import numpy_force as numpy


@bhary.fix_biclass_wrapper
def sum(a, axis=None, dtype=None, out=None):
    """
    Sum of array elements over a given axis.

    Parameters
    ----------
    a : array_like
        Elements to sum.
    axis : None or int or tuple of ints, optional
        Axis or axes along which a sum is performed.
        The default (`axis` = `None`) is perform a sum over all
        the dimensions of the input array. `axis` may be negative, in
        which case it counts from the last to the first axis.

        .. versionadded:: 1.7.0

        If this is a tuple of ints, a sum is performed on multiple
        axes, instead of a single axis or all the axes as before.
    dtype : dtype, optional
        The type of the returned array and of the accumulator in which the
        elements are summed.  The dtype of `a` is used by default unless `a`
        has an integer dtype of less precision than the default platform
        integer.  In that case, if `a` is signed then the platform integer
        is used while if `a` is unsigned then an unsigned integer of the
        same precision as the platform integer is used.
    out : ndarray, optional
        Array into which the output is placed.  By default, a new array is
        created.  If `out` is given, it must be of the appropriate shape
        (the shape of `a` with `axis` removed, i.e.,
        ``numpy.delete(a.shape, axis)``).  Its type is preserved. See
        `doc.ufuncs` (Section "Output arguments") for more details.

    Returns
    -------
    sum_along_axis : ndarray
        An array with the same shape as `a`, with the specified
        axis removed.   If `a` is a 0-d array, or if `axis` is None, a scalar
        is returned.  If an output array is specified, a reference to
        `out` is returned.

    See Also
    --------
    ndarray.sum : Equivalent method.

    cumsum : Cumulative sum of array elements.

    trapz : Integration of array values using the composite trapezoidal rule.

    mean, average

    Notes
    -----
    Arithmetic is modular when using integer types, and no error is
    raised on overflow.

    Examples
    --------
    >>> np.sum([0.5, 1.5])
    2.0
    >>> np.sum([0.5, 0.7, 0.2, 1.5], dtype=np.int32)
    1
    >>> np.sum([[0, 1], [0, 5]])
    6
    >>> np.sum([[0, 1], [0, 5]], axis=0)
    array([0, 6])
    >>> np.sum([[0, 1], [0, 5]], axis=1)
    array([1, 5])

    If the accumulator is too small, overflow occurs:

    >>> np.ones(128, dtype=np.int8).sum(dtype=np.int8)
    -128

    """

    if not bhary.check(a) and not bhary.check(out):
        return numpy.sum(a, axis=axis, dtype=dtype, out=out)
    else:
        if dtype is not None:
            a = array_create.array(a, dtype=dtype)
        return ufuncs.add.reduce(a, axis=axis, out=out)


@bhary.fix_biclass_wrapper
def prod(a, axis=None, dtype=None, out=None):
    """
    Product of array elements over a given axis.

    Parameters
    ----------
    a : array_like
        Elements to multiply.
    axis : None or int or tuple of ints, optional
        Axis or axes along which a multiply is performed.
        The default (`axis` = `None`) is perform a multiply over all
        the dimensions of the input array. `axis` may be negative, in
        which case it counts from the last to the first axis.

        .. versionadded:: 1.7.0

        If this is a tuple of ints, a multiply is performed on multiple
        axes, instead of a single axis or all the axes as before.
    dtype : dtype, optional
        The type of the returned array and of the accumulator in which the
        elements are summed.  The dtype of `a` is used by default unless `a`
        has an integer dtype of less precision than the default platform
        integer.  In that case, if `a` is signed then the platform integer
        is used while if `a` is unsigned then an unsigned integer of the
        same precision as the platform integer is used.
    out : ndarray, optional
        Array into which the output is placed.  By default, a new array is
        created.  If `out` is given, it must be of the appropriate shape
        (the shape of `a` with `axis` removed, i.e.,
        ``numpy.delete(a.shape, axis)``).  Its type is preserved. See
        `doc.ufuncs` (Section "Output arguments") for more details.

    Returns
    -------
    protuct_along_axis : ndarray
        An array with the same shape as `a`, with the specified
        axis removed.   If `a` is a 0-d array, or if `axis` is None, a scalar
        is returned.  If an output array is specified, a reference to
        `out` is returned.

    Examples
    --------
    >>> np.prod([0.5, 1.5])
    2.0
    >>> np.prod([0.5, 0.7, 0.2, 1.5], dtype=np.int32)
    1
    >>> np.prod([[0, 1], [0, 5]])
    6
    >>> np.prod([[0, 1], [0, 5]], axis=0)
    array([0, 6])
    >>> np.prod([[0, 1], [0, 5]], axis=1)
    array([1, 5])

    If the accumulator is too small, overflow occurs:

    >>> np.ones(128, dtype=np.int8).prod(dtype=np.int8)
    -128

    """

    if not bhary.check(a) and not bhary.check(out):
        return numpy.prod(a, axis=axis, dtype=dtype, out=out)
    else:
        if dtype is not None:
            a = array_create.array(a, dtype=dtype)
        return ufuncs.multiply.reduce(a, axis=axis, out=out)


@bhary.fix_biclass_wrapper
def max(a, axis=None, out=None):
    """
    Return the maximum of an array or maximum along an axis.

    Parameters
    ----------
    a : array_like
        Input data.
    axis : None or int or tuple of ints, optional
        Axis or axes along which a max is performed.
        The default (`axis` = `None`) is perform a max over all
        the dimensions of the input array. `axis` may be negative, in
        which case it counts from the last to the first axis.
        If this is a tuple of ints, a max is performed on multiple
        axes, instead of a single axis or all the axes as before.
    out : ndarray, optional
        Alternative output array in which to place the result.  Must
        be of the same shape and buffer length as the expected output.
        See `doc.ufuncs` (Section "Output arguments") for more details.

    Returns
    -------
    max : ndarray or scalar
        Maximum of `a`. If `axis` is None, the result is a scalar value.
        If `axis` is given, the result is an array of dimension
        ``a.ndim - 1``.

    See Also
    --------
    min :
        The minimum value of an array along a given axis, propagating any NaNs.
    nanmax :
        The maximum value of an array along a given axis, ignoring any NaNs.
    maximum :
        Element-wise maximum of two arrays, propagating any NaNs.
    fmax :
        Element-wise maximum of two arrays, ignoring any NaNs.
    argmax :
        Return the indices of the maximum values.

    nanmin, minimum, fmin

    Notes
    -----
    NaN values are propagated, that is if at least one item is NaN, the
    corresponding max value will be NaN as well. To ignore NaN values
    (MATLAB behavior), please use nanmax.

    Don't use `max` for element-wise comparison of 2 arrays; when
    ``a.shape[0]`` is 2, ``maximum(a[0], a[1])`` is faster than
    ``max(a, axis=0)``.

    Examples
    --------
    >>> a = np.arange(4).reshape((2,2))
    >>> a
    array([[0, 1],
           [2, 3]])
    >>> np.max(a)           # Maximum of the flattened array
    3
    >>> np.max(a, axis=0)   # Maxima along the first axis
    array([2, 3])
    >>> np.max(a, axis=1)   # Maxima along the second axis
    array([1, 3])

    >>> b = np.arange(5, dtype=np.float)
    >>> b[2] = np.NaN
    >>> np.max(b)
    nan
    >>> np.nanmax(b)
    4.0

    """

    if not bhary.check(a) and not bhary.check(out):
        return numpy.max(a, axis=axis, out=out)  # NumPy 1.6 doesn't support axis=None
    else:
        return ufuncs.maximum.reduce(a, axis=axis, out=out)


@bhary.fix_biclass_wrapper
def min(a, axis=None, out=None):
    """
    Return the minimum of an array or minimum along an axis.

    Parameters
    ----------
    a : array_like
        Input data.
    axis : None or int or tuple of ints, optional
        Axis or axes along which a min is performed.
        The default (`axis` = `None`) is perform a min over all
        the dimensions of the input array. `axis` may be negative, in
        which case it counts from the last to the first axis.
        If this is a tuple of ints, a min is performed on multiple
        axes, instead of a single axis or all the axes as before.
    out : ndarray, optional
        Alternative output array in which to place the result.  Must
        be of the same shape and buffer length as the expected output.
        See `doc.ufuncs` (Section "Output arguments") for more details.

    Returns
    -------
    min : ndarray or scalar
        minimum of `a`. If `axis` is None, the result is a scalar value.
        If `axis` is given, the result is an array of dimension
        ``a.ndim - 1``.

    See Also
    --------
    min :
        The minimum value of an array along a given axis, propagating any NaNs.
    nanmin :
        The minimum value of an array along a given axis, ignoring any NaNs.
    minimum :
        Element-wise minimum of two arrays, propagating any NaNs.
    fmin :
        Element-wise minimum of two arrays, ignoring any NaNs.
    argmin :
        Return the indices of the minimum values.

    nanmin, minimum, fmin

    Notes
    -----
    NaN values are propagated, that is if at least one item is NaN, the
    corresponding min value will be NaN as well. To ignore NaN values
    (MATLAB behavior), please use nanmin.

    Don't use `min` for element-wise comparison of 2 arrays; when
    ``a.shape[0]`` is 2, ``minimum(a[0], a[1])`` is faster than
    ``min(a, axis=0)``.

    Examples
    --------
    >>> a = np.arange(4).reshape((2,2))
    >>> a
    array([[0, 1],
           [2, 3]])
    >>> np.min(a)           # minimum of the flattened array
    3
    >>> np.min(a, axis=0)   # minima along the first axis
    array([2, 3])
    >>> np.min(a, axis=1)   # minima along the second axis
    array([1, 3])

    >>> b = np.arange(5, dtype=np.float)
    >>> b[2] = np.NaN
    >>> np.min(b)
    nan
    >>> np.nanmin(b)
    4.0

    """

    if not bhary.check(a) and not bhary.check(out):
        return numpy.min(a, axis=axis, out=out)  # NumPy 1.6 doesn't support axis=None
    else:
        return ufuncs.minimum.reduce(a, axis=axis, out=out)


@bhary.fix_biclass_wrapper
def any(a, axis=None, out=None, keepdims=None):
    if not bhary.check(a) and not bhary.check(out):
        return numpy.any(a, axis=axis, out=out)
    else:
        return ufuncs.logical_or.reduce(a.astype(bool), axis=axis, out=out)


@bhary.fix_biclass_wrapper
def all(a, axis=None, out=None, keepdims=None):
    if not bhary.check(a) and not bhary.check(out):
        return numpy.all(a, axis=axis, out=out)
    else:
        return ufuncs.logical_and.reduce(a.astype(bool), axis=axis, out=out)


@bhary.fix_biclass_wrapper
def argmax(a, axis=None, out=None):
    """
    Returns the indices of the maximum values along an axis.
    Parameters
    ----------
    a : array_like
        Input array.
    axis : int, optional
        By default, the index is into the flattened array, otherwise
        along the specified axis.
    out : array, optional
        If provided, the result will be inserted into this array. It should
        be of the appropriate shape and dtype.
    Returns
    -------
    index_array : ndarray of ints
        Array of indices into the array. It has the same shape as `a.shape`
        with the dimension along `axis` removed.
    See Also
    --------
    ndarray.argmax, argmin
    amax : The maximum value along a given axis.
    unravel_index : Convert a flat index into an index tuple.
    Notes
    -----
    In case of multiple occurrences of the maximum values, the indices
    corresponding to the first occurrence are returned.
    Examples
    --------
    >>> a = np.arange(6).reshape(2,3)
    >>> a
    array([[0, 1, 2],
           [3, 4, 5]])
    >>> np.argmax(a)
    5
    >>> np.argmax(a, axis=0)
    array([1, 1, 1])
    >>> np.argmax(a, axis=1)
    array([2, 2])
    >>> b = np.arange(6)
    >>> b[1] = 5
    >>> b
    array([0, 5, 2, 3, 4, 5])
    >>> np.argmax(b) # Only the first occurrence is returned.
    1
    """

    if not bhary.check(a):
        return numpy.argmax(a, axis=axis, out=out)

    if axis is None or (a.ndim == 1 and axis == 0):
        a = array_manipulation.flatten(a, always_copy=False)
        ret = reorganization.flatnonzero(a == max(a))[0]
    else:
        warnings.warn("Bohrium does not support the 'axis' argument, "
                      "it will be handled by the original NumPy.", UserWarning, 2)
        return numpy.argmax(a.copy2numpy(), axis=axis)

    if out is None:
        return ret
    else:
        out[...] = ret
        return out


@bhary.fix_biclass_wrapper
def argmin(a, axis=None, out=None):
    """
    Returns the indices of the minimum values along an axis.
    Parameters
    ----------
    a : array_like
        Input array.
    axis : int, optional
        By default, the index is into the flattened array, otherwise
        along the specified axis.
    out : array, optional
        If provided, the result will be inserted into this array. It should
        be of the appropriate shape and dtype.
    Returns
    -------
    index_array : ndarray of ints
        Array of indices into the array. It has the same shape as `a.shape`
        with the dimension along `axis` removed.
    See Also
    --------
    ndarray.argmin, argmax
    amin : The minimum value along a given axis.
    unravel_index : Convert a flat index into an index tuple.
    Notes
    -----
    In case of multiple occurrences of the minimum values, the indices
    corresponding to the first occurrence are returned.
    Examples
    --------
    >>> a = np.arange(6).reshape(2,3)
    >>> a
    array([[0, 1, 2],
           [3, 4, 5]])
    >>> np.argmin(a)
    0
    >>> np.argmin(a, axis=0)
    array([0, 0, 0])
    >>> np.argmin(a, axis=1)
    array([0, 0])
    >>> b = np.arange(6)
    >>> b[4] = 0
    >>> b
    array([0, 1, 2, 3, 0, 5])
    >>> np.argmin(b) # Only the first occurrence is returned.
    0
    """

    if not bhary.check(a):
        return numpy.argmin(a, axis=axis, out=out)

    if axis is None or (a.ndim == 1 and axis == 0):
        a = array_manipulation.flatten(a, always_copy=False)
        ret = reorganization.flatnonzero(a == min(a))[0]
    else:
        warnings.warn("Bohrium does not support the 'axis' argument, "
                      "it will be handled by the original NumPy.", UserWarning, 2)
        return numpy.argmin(a.copy2numpy(), axis=axis)

    if out is None:
        return ret
    else:
        out[...] = ret
        return out


def mean(a, axis=None, dtype=None, out=None):
    """
    Compute the arithmetic mean along the specified axis.
    Returns the average of the array elements.  The average is taken over
    the flattened array by default, otherwise over the specified axis.
    `float64` intermediate and return values are used for integer inputs.
    Parameters
    ----------
    a : array_like
        Array containing numbers whose mean is desired. If `a` is not an
        array, a conversion is attempted.
    axis : None or int or tuple of ints, optional
        Axis or axes along which the means are computed. The default is to
        compute the mean of the flattened array.
        .. versionadded:: 1.7.0
        If this is a tuple of ints, a mean is performed over multiple axes,
        instead of a single axis or all the axes as before.
    dtype : data-type, optional
        Type to use in computing the mean.  For integer inputs, the default
        is `float64`; for floating point inputs, it is the same as the
        input dtype.
    out : ndarray, optional
        Alternate output array in which to place the result.  The default
        is ``None``; if provided, it must have the same shape as the
        expected output, but the type will be cast if necessary.
        See `doc.ufuncs` for details.

    Returns
    -------
    m : ndarray, see dtype parameter above
        If `out=None`, returns a new array containing the mean values,
        otherwise a reference to the output array is returned.
    See Also
    --------
    average : Weighted average
    std, var, nanmean, nanstd, nanvar
    Notes
    -----
    The arithmetic mean is the sum of the elements along the axis divided
    by the number of elements.
    Note that for floating-point input, the mean is computed using the
    same precision the input has.  Depending on the input data, this can
    cause the results to be inaccurate, especially for `float32` (see
    example below).  Specifying a higher-precision accumulator using the
    `dtype` keyword can alleviate this issue.
    By default, `float16` results are computed using `float32` intermediates
    for extra precision.
    Examples
    --------
    >>> a = np.array([[1, 2], [3, 4]])
    >>> np.mean(a)
    2.5
    >>> np.mean(a, axis=0)
    array([ 2.,  3.])
    >>> np.mean(a, axis=1)
    array([ 1.5,  3.5])
    In single precision, `mean` can be inaccurate:
    >>> a = np.zeros((2, 512*512), dtype=np.float32)
    >>> a[0, :] = 1.0
    >>> a[1, :] = 0.1
    >>> np.mean(a)
    0.54999924
    Computing the mean in float64 is more accurate:
    >>> np.mean(a, dtype=np.float64)
    0.55000000074505806
    """

    def _count_reduce_items(arr, axis):
        if axis is None:
            axis = tuple(range(arr.ndim))
        if not isinstance(axis, tuple):
            axis = (axis,)
        items = 1
        for ax in axis:
            items *= arr.shape[ax]
        return items

    def _mean(a, axis=None, dtype=None, out=None):
        arr = array_create.array(a)
        is_float16_result = False
        rcount = _count_reduce_items(arr, axis)
        # Make this warning show up first
        if rcount == 0:
            warnings.warn("Mean of empty slice.", RuntimeWarning, stacklevel=2)

        # Cast bool, unsigned int, and int to float64 by default
        if dtype is None:
            if issubclass(arr.dtype.type, (numpy.integer, numpy.bool_)):
                dtype = numpy.dtype('f8')
            elif issubclass(arr.dtype.type, numpy.float16):
                dtype = numpy.dtype('f4')
                is_float16_result = True

        ret = sum(arr, axis, dtype, out)
        if isinstance(ret, numpy.ndarray):
            ret = ufuncs.true_divide(ret, rcount, out=ret)
            if is_float16_result and out is None:
                ret = a.dtype.type(ret)
        elif hasattr(ret, 'dtype'):
            if is_float16_result:
                ret = a.dtype.type(ret / rcount)
            else:
                ret = ret.dtype.type(ret / rcount)
        else:
            ret = ret / rcount
        return ret
    return _mean(a, axis=axis, dtype=dtype, out=out)


average = mean
